/*****************************************************************************/
/*
                                  data.c

COPYRIGHT
---------
Copyright (C) 2008-2011 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it under the
conditions of the GNU GENERAL PUBLIC LICENSE, version 3, or any later version.
http://www.gnu.org/licenses/gpl.txt


MEMORY ALLOCATION
-----------------
If the cache is enabled and the file can be cached then memory is allocated
using calloc() and therefore persists between requests.  If the file cannot be
cached then memory is allocated using CgiLibVeeMemCalloc() and is therefore
automatically disposed of at the end of the request.


VERSION HISTORY
---------------
24-MAR-2008  MGD  initial
*/

/*****************************************************************************/

#ifdef VWCMS_VMS_V62
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60200000
#undef __CRTL_VER
#define __CRTL_VER 60200000
#endif

#ifndef __VAX
#   pragma nomember_alignment
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stat.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <types.h>

/* VMS related header files */
#include <devdef.h>
#include "dvidef.h"
#include "fabdef.h"
#include "ssdef.h"
#include "stsdef.h"
#include "rmsdef.h"

/* application header file */
#include "cgilib.h"
#include "config.h"
#include "data.h"
#include "vwcms.h"
#include "other.h"

#ifndef uchar
#define uchar unsigned char
#endif
#ifndef ushort
#define ushort unsigned short
#endif
#ifndef uint
#define uint unsigned int
#endif
#ifndef ulong
#define ulong unsigned long
#endif

#define FI_LI "DATA", __LINE__

/* global storage */

int  DataCacheCount,
     DataCacheMax,
     DataCacheSizeMax,
     DataNameMax,
     DataVolumeOds;

DATA_CACHE *DataCache;

/* prototypes */

int sys$fao (__unknown_params);
int sys$getdviw (__unknown_params);

/* external storage */

extern BOOL  Debug,
             IsCgiPlus,
             WatchEnabled;

extern char  *CgiPathInfoPtr,
             *CgiServerNamePtr;

extern char  LoginUserName[],
             SoftwareId[],
             Utility[];

extern CONFIG_DATA  VWcmsConfig;

/*****************************************************************************/
/*
*/

void DataCacheInit ()

{
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DataCacheInit()\n");

   if (!IsCgiPlus) return;

   cptr = TrnLnm ("VWCMS_DATA_CACHE_MAX", 0, 0);
   if (cptr) DataCacheMax = atoi(cptr);
   if (DataCacheMax <= 0) DataCacheMax = DATA_CACHE_MAX_DEFAULT;

   cptr = TrnLnm ("VWCMS_DATA_CACHE_SIZE_MAX", 0, 0);
   if (cptr) DataCacheSizeMax = atoi(cptr);
   if (DataCacheSizeMax <= 0) DataCacheSizeMax = DATA_CACHE_SIZE_MAX_DEFAULT;

   DataCache = (DATA_CACHE*)calloc (DataCacheMax, sizeof(DATA_CACHE));
   if (!DataCache) ErrorExit (vaxc$errno, FI_LI);
}

/****************************************************************************/
/*
Read the file contents specified by 'FileName' into memory, set the pointer
at 'FileTextPtr' to the contents and the file size at 'FileSizePtr'.

Returns a VMS status value that should be checked.

Cache file content if CGIplus and cache enabled.
*/ 

int DataReadFileIntoMemory
(
char *FileName,
char **FileTextPtr,
int *FileSizePtr,
BOOL UseCache
)
{
   static int  DataCacheReuse;
   static char  *BufferPtr;

   int  retval, status,
        Bytes,
        BytesRemaining,
        BufferCount,
        CacheIndex;
   char  *cptr, *sptr,
         *LinePtr;
   FILE  *FilePtr;
   DATA_CACHE  *dcptr;
   stat_t  StatBuffer;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DataReadFileIntoMemory() |%s|\n", FileName);

   if (FileTextPtr != NULL) *FileTextPtr = NULL;
   if (FileSizePtr != NULL) *FileSizePtr = 0;

   if (stat (FileName, &StatBuffer) < 0)
   {
      status = vaxc$errno;
      if (Debug) fprintf (stdout, "stat() %%X%08.08X\n", status);
      return (status);
   }

   if (!DataCacheMax) UseCache = FALSE;

   Bytes = StatBuffer.st_size;
   if (Debug) fprintf (stdout, "%d bytes\n", Bytes);

   /* in kilobytes */
   if (Bytes / 1024 > DataCacheSizeMax) UseCache = FALSE;

   if (UseCache)
   {
      /***************/
      /* check cache */
      /***************/

      for (CacheIndex = 0; CacheIndex < DataCacheCount; CacheIndex++)
      {
         dcptr = &DataCache[CacheIndex];
         if (Debug)
            fprintf (stdout, "%d |%s|%d|%s|%d| %d\n",
                     CacheIndex, FileName, StatBuffer.st_mtime,
                     dcptr->FileName, dcptr->mtime, dcptr->ContentLength);
         if (strsame (dcptr->FileName, FileName, -1)) break;
      }

      if (CacheIndex < DataCacheCount)
      {
         if (dcptr->mtime == StatBuffer.st_mtime)
         {
            /********/
            /* hit! */
            /********/

            dcptr->HitCount++;
            if (FileTextPtr) *FileTextPtr = dcptr->ContentPtr;
            if (FileSizePtr) *FileSizePtr = dcptr->ContentLength;
            if (Debug) fprintf (stdout, "%s", *FileTextPtr);
            return (SS$_NORMAL);
         }

         /***************/
         /* flush entry */
         /***************/

         if (dcptr->ContentPtr)
         {
            free (dcptr->ContentPtr);
            dcptr->ContentPtr = NULL;
            dcptr->FileName[0] = '\0';
         }
      }
      else
         CacheIndex = -1;
   }

   /*************/
   /* read file */
   /*************/

   if (Debug) fprintf (stdout, "st_fab_rfm: %d\n", StatBuffer.st_fab_rfm);

   if (StatBuffer.st_fab_rfm == FAB$C_VAR ||
       StatBuffer.st_fab_rfm == FAB$C_VFC)
      FilePtr = fopen (FileName, "r", "shr=put");
   else
      FilePtr = fopen (FileName, "r", "shr=put", "ctx=bin");
   if (FilePtr == NULL)
   {
      status = vaxc$errno;
      if (Debug) fprintf (stdout, "fopen() %%X%08.08X\n", status);
      return (status);
   }

   /* a little margin for error ;^) */
   if (UseCache)
      BufferPtr = calloc (Bytes+32, 1);
   else
      BufferPtr = CgiLibVeeMemCalloc (Bytes+32);
   if (!BufferPtr) ErrorExit (vaxc$errno, FI_LI);
   BufferCount = 0;

   if (StatBuffer.st_fab_rfm == FAB$C_VAR ||
       StatBuffer.st_fab_rfm == FAB$C_VFC)
   {
      BytesRemaining = Bytes;
      LinePtr = BufferPtr;
      while (fgets (LinePtr, BytesRemaining, FilePtr) != NULL)
      {
         /** if (Debug) fprintf (stdout, "|%s|\n", LinePtr); **/
         if (!*LinePtr) break;
         for (cptr = LinePtr; *cptr; cptr++);
         BufferCount += cptr - LinePtr;
         BytesRemaining -= cptr - LinePtr;
         LinePtr = cptr;
      }
   }
   else
   {
      retval = fread (BufferPtr, Bytes, 1, FilePtr);
      if (retval == 1) BufferCount = Bytes;
   }

   fclose (FilePtr);

   if (Debug) fprintf (stdout, "%d |%s|\n", BufferCount, BufferPtr);

   if (FileTextPtr) *FileTextPtr = BufferPtr;
   if (FileSizePtr) *FileSizePtr = BufferCount;

   /* if no cache then quit here */
   if (!UseCache) return (SS$_NORMAL);

   /*****************/
   /* cache content */
   /*****************/

   if (CacheIndex < 0)
   {
      /* previously uncached content */
      if (DataCacheCount < DataCacheMax)
      {
         CacheIndex = DataCacheCount++;
         dcptr = &DataCache[CacheIndex];
      }
      else
      {
         /* need to reuse a cache entry - unsophisticated FIFO! */
         if (DataCacheReuse >= DataCacheMax) DataCacheReuse = 0;
         CacheIndex = DataCacheReuse++;
         dcptr = &DataCache[CacheIndex];
         dcptr->ReusedCount++;
         if (dcptr->ContentPtr)
         {
            free (dcptr->ContentPtr);
            dcptr->ContentPtr = NULL;
            dcptr->FileName[0] = '\0';
         }
      }
   }

   dcptr->ContentPtr = BufferPtr;
   dcptr->ContentLength = BufferCount;
   dcptr->mtime = StatBuffer.st_mtime;
   dcptr->HitCount = 0;
   strcpy (dcptr->FileName, FileName);

   return (SS$_NORMAL);
}

/****************************************************************************/
/*
Remove the specified file from the data cache.
*/ 

BOOL DataCacheFlush (char *FileName)

{
   int  CacheIndex;
   DATA_CACHE  *dcptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DataCacheFlush() |%s|\n", FileName);

   if (!DataCacheMax) return (FALSE);

   for (CacheIndex = 0; CacheIndex < DataCacheCount; CacheIndex++)
   {
      dcptr = &DataCache[CacheIndex];
      if (Debug)
         fprintf (stdout, "%d |%s|%s|\n",
                  CacheIndex, FileName, dcptr->FileName);
      if (strsame (dcptr->FileName, FileName, -1)) break;
   }
   if (CacheIndex >= DataCacheCount) return (FALSE);

   free (dcptr->ContentPtr);
   dcptr->ContentPtr = NULL;
   dcptr->FileName[0] = '\0';

   return (TRUE);
}

/****************************************************************************/
/*
Get the last-modified time for the specified 'FileName'.
Returns zero if there was an error.
*/ 

int DataFileMtime
(
char *FileName,
BOOL UseCache
)
{
   int  retval,
        CacheIndex;
   DATA_CACHE  *dcptr;
   stat_t  StatBuffer;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DataFileMtime() |%s|\n", FileName);

   if (UseCache)
   {
      for (CacheIndex = 0; CacheIndex < DataCacheCount; CacheIndex++)
      {
         dcptr = &DataCache[CacheIndex];
         if (Debug)
            fprintf (stdout, "%d |%s|%s|\n",
                     CacheIndex, FileName, dcptr->FileName);
         if (strsame (dcptr->FileName, FileName, -1)) break;
      }
      if (CacheIndex < DataCacheCount) return (dcptr->mtime);
   }

   if (stat (FileName, &StatBuffer) >= 0) return (StatBuffer.st_mtime);
   if (Debug) fprintf (stdout, "stat() %%X%08.08X\n", vaxc$errno);

   return (0);
}

/****************************************************************************/
/*
Get the volume's underlying file system (ODS-2 or ODS-5), by sys$getdvi() the
ACPTYPE for the device.  Set the required global storage and return 5 if ODS-5
(EFS), 2 if ODS-2, zero on error or anything else.  The logical name allows an
ODS-2 volume naming scheme to be forced to be used on an ODS-5 volume.
*/ 

/* if pre-7.2 Alpha then define this */
#ifndef DVI$C_ACP_F11V5
#define DVI$C_ACP_F11V5 11
#endif

int DataVolumeStruct (char *DeviceName)

{
   static int  DevOds;
   static unsigned long  DevAcpType,
                         DevChar;
   static $DESCRIPTOR (DevNameDsc, "");
   static VMS_ITEM_LIST3  AcpTypeItemList [] = 
   {
      { sizeof(DevAcpType), DVI$_ACPTYPE, &DevAcpType, 0 },
      { sizeof(DevChar), DVI$_DEVCHAR, &DevChar, 0 },
      { 0, 0, 0, 0 }
   };

   int  idx, status;
   char  *cptr;
   struct {
      unsigned short  Status;
      unsigned short  Count;
      unsigned long  Unused;
   } IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "DataVolumeStruct()\n");

   for (cptr = DeviceName; *cptr && *cptr != ':'; cptr++);
   /* if we can't find a device in it */
   if (!*cptr) return (0);

   DevNameDsc.dsc$a_pointer = DeviceName;
   DevNameDsc.dsc$w_length = cptr - DeviceName;

   DevChar = DevAcpType = 0;

   status = sys$getdviw (0, 0, &DevNameDsc, &AcpTypeItemList,
                         &IOsb, 0, 0, 0);
   if (Debug)
      fprintf (stdout, "sys$getdviw() |%*s| %%X%08.08X %%X%08.08X %d\n",
               DevNameDsc.dsc$w_length, DevNameDsc.dsc$a_pointer,
               status, IOsb.Status, DevAcpType);

   if (VMSok (status)) status = IOsb.Status;
   if (VMSnok (status)) ErrorExit (status, FI_LI); 

   /* if volume not mounted */
   if (!(DevChar & DEV$M_MNT)) return (0);
                             
   if (DevAcpType == DVI$C_ACP_F11V5) DataVolumeOds = 5;
   if (DevAcpType == DVI$C_ACP_F11V2) DataVolumeOds = 2;

   if (!DataVolumeOds) ErrorExit (SS$_BUGCHECK, FI_LI); 

   return (DataVolumeOds);
}

/*****************************************************************************/

