/*****************************************************************************/
/*
                                 form.c

User form via email or POST-processing spawn.

$$FORM$action=<keyword>=<parameter>$$
$$FORM$area=<name>$size=<row>x<col>$value=<string>$$
$$FORM$check=<name>$one$two$three$$
$$FORM$format=<keyword>$$
$$FORM$insert=<something>$$
$$FORM$password=<name>$size=<integer>$value=<string>$$
$$FORM$radio=<name>$one$two$three$$
$$FORM$reset=<string>$$
$$FORM$set=<legend>$$
$$FORM$spawned=<keyword>$$
$$FORM$submit=<string>$$
$$FORM$test$$
$$FORM$text=<name>$size=<integer>$value=<string>$$
$$FORM$transport=<string>$$

Form elements (and appropriate values) are disabled by prefixing them with '!'.

  $$FORM$!check=<name>$one$two$three$$
  
Appropriate values are selected/checked by prefixing them with '`' (back-tick).

  $$FORM$check=<name>$`one$two$`three$$

The special '#' makes an element 'standalone' for example purposes.

$$FORM$test$$ disables posting of the form and displays what would be actioned
instead.  $$FORM$!test$$ disables this development feature.


FORM DATA FORMAT
----------------
Form data formats are:
1) RAW - just the POSTed body.
2) PLAIN - a readable, plain-text rendition of the body.
3) HTML - a readable, HTMl table containing the body data.
4) XML - generates a fragment of XML ready to be inserted into a DOCTYPE.

If a page with a $$FORM$$ directive also has a $$TICKET$$ directive then one of
the form text fields must contain the ticket string.  This attempts to impede
SPAMbot submissions.  See TicketFile() for implementation comment.


FORM SPAWN
----------
The $$FORM$spawn=<procedure>$$ directive provides the name of a DCL command
procedure located in the site directory.  So if the directive is

  $$FORM$spawn=just_an_example$$

then VWcms will spawn a subprocess to execute

  @device:[directory]_JUST_AN_EXAMPLE.COM

with the first parameter (P1) containing the name of the temporary file
containing the form data in the format requested (raw, plain, HTML, etc.) and
anything following the spawn ("just_an_example") from the directive appended
to that as DCL (and requiring all that DCL demands).

Any output generated by the spawn is subsequently and optionally available for
inclusion in a page using the $$FORM$spawned=plain$$ (HTML-escaped) or the
$$FORM$spawned=html$$ directives anywhere in the content.

The management interface will reset all _*.COM to owner-only protections and so
any command procedure in the site directory should have an ACL allowing the
scripting account access; something like

   $ DIR /SECURITY WASD_ROOT:[SRC.VWCMS.VDUB]_EXAMPLEFORMSPAWN.COM;1

   Directory WASD_ROOT:[SRC.VWCMS.VDUB]

   _EXAMPLEFORMSPAWN.COM;1
                         0.50KB  19-DEC-2009 22:14:02.63  [SYSTEM]
           (RWED,RWED,,)
             (IDENTIFIER=WASD_HTTP_NOBODY,ACCESS=READ+EXECUTE)
             (IDENTIFIER=*,ACCESS=NONE)

   Total of 1 file, 0.50KB



COPYRIGHT
---------
Copyright (C) 2008-2011 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it under the
conditions of the GNU GENERAL PUBLIC LICENSE, version 3, or any later version.
http://www.gnu.org/licenses/gpl.txt


VERSION HISTORY
---------------
18-JUL-2011  MGD  Google reCAPTCHA
04-MAY-2011  MGD  XHTMLified some fugitive HTML
26-FEB-2011  MGD  configuration mail transport
                  bugfix; FormSubmitSpawn() fopen/close/delete than "fop=dlt"
02-DEC-2009  MGD  initial
*/

/*****************************************************************************/

#ifdef VWCMS_VMS_V62
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60200000
#undef __CRTL_VER
#define __CRTL_VER 60200000
#endif

#ifndef __VAX
#   pragma nomember_alignment
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <locale.h>
#include <stat.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <types.h>
#include <time.h>
#include <unistd.h>
#include <unixio.h>

/* VMS related header files */
#include "descrip.h"
#include "rmsdef.h"
#include "ssdef.h"
#include "stsdef.h"
#include "starlet.h"

/* application header file */
#include "cgilib.h"
#include "config.h"
#include "comment.h"
#include "content.h"
#include "cookie.h"
#include "data.h"
#include "form.h"
#include "other.h"
#include "qanda.h"
#include "recaptcha.h"
#include "sendmail.h"
#include "template.h"
#include "ticket.h"
#include "vwcms.h"

#define FI_LI "FORM", __LINE__

#define FORM_FIELD_ACTION      1
#define FORM_FIELD_AREA        2
#define FORM_FIELD_CHECK       3
#define FORM_FIELD_INSERT      4
#define FORM_FIELD_RADIO       5
#define FORM_FIELD_RESET       6
#define FORM_FIELD_SELECT      7
#define FORM_FIELD_SET         8
#define FORM_FIELD_SUBMIT      9
#define FORM_FIELD_TEXT       10
#define FORM_FIELD_PASSWORD   11

#define FORM_DATA_FORMAT_RAW   1
#define FORM_DATA_FORMAT_PLAIN 2
#define FORM_DATA_FORMAT_HTML  3
#define FORM_DATA_FORMAT_XML   4

/* prototypes */

/* global storage */

BOOL  FormSubmitDone;

char  *FormSpawnOutputPtr;

/* prototypes */

/* external storage */

extern BOOL  Debug,
             QandaAfter,
             QandaBefore,
             ReCaptchaAfter,
             ReCaptchaBefore,
             RequestMethodPOST,
             TicketOnPage,
             WatchEnabled;

extern int  CgiPlusUsageCount,
            CurrentSecondsTime,
            PostBufferCount;

extern unsigned long  CurrentVmsBinTime[];
extern unsigned short  CurrentVmsNumTime[];

extern char  *CgiRequestUriPtr,
             *PostBufferPtr;

extern char  ContentSectionName[],
             SoftwareId[],
             Utility[];

extern CONTENT_DATA  *ContentSectionPtr;

extern CONFIG_DATA  VWcmsConfig;

/*****************************************************************************/
/*
Called by a $$FORM$..$$ directive in TemplateLoad() or TemplateContent(). 
Places processed form directives into the displayed content.
*/

void FormTemplate (char *SubsArg)

{
   static BOOL  FormDisabled,
                FormFieldSet,
                FormOpen;
   static int  DataFormat,
               UsageCount;

   BOOL  AlignVert = FALSE,
         FieldDisabled = FALSE,
         FieldExample = FALSE,
         FieldAttention = FALSE,
         FieldRequired = FALSE,
         ValueChecked,
         ValueDisabled;
   int  CheckCount,
        FieldSize = 0,
        FieldWant = 0,
        FormCount,
        HitCount,
        SelectSize = 0,
        TextRows = 0,
        TextCols = 0,
        ThisField = 0;
   char  *aptr, *cptr, *sptr, *zptr,
         *ValidPtr = NULL,
         *ValuePtr = NULL;
   char  FormAction [256] = "",
         FormAlign [32] = "",
         FormFormat [32] = "",
         FormGoto [256] = "",
         FormSize [32] = "",
         FieldName [256] = "",
         FieldValue [1024] = "";

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "FormTemplate() |%s|\n", SubsArg);

   if (CgiPlusUsageCount != UsageCount)
   {
      UsageCount = CgiPlusUsageCount;
      FormDisabled = FormFieldSet = FormOpen = FormSubmitDone = FALSE;
      DataFormat = FormCount = 0;
   }

   if (!SubsArg || !*SubsArg)
   {
      /***************/
      /* end of form */
      /***************/

      if (FormFieldSet) fputs ("</fieldset>\n", stdout);
      if (FormOpen) fputs ("</form>\n", stdout);
      FormFieldSet = FormOpen = FALSE;
      return;
   }

   if (*(cptr = SubsArg) == '$') cptr++;

   if (strsame (cptr, "test", -1))
   {
      /*****************/
      /* posted output */
      /*****************/

      /* special case */
      if (!RequestMethodPOST) return;
      sptr = FormFormatData (0, NULL, NULL, NULL);
      if (!sptr)
      {
         TemplateAdvise ("E$$FORM$!AZ$$ ?", cptr);
         return;
      }

      if (DataFormat == FORM_DATA_FORMAT_RAW)
      {
         fputs ("<div class=\"vwcms_form_test\"><pre>", stdout);
         /* should really HTML-escape in the dump but all too hard! */
         WatchDump (sptr, strlen(sptr));
         fputs ("</pre></div>", stdout);
      }
      else
      if (DataFormat == FORM_DATA_FORMAT_HTML)
         fprintf (stdout, "<div class=\"vwcms_form_test\">%s</div>", sptr);
      else
         fprintf (stdout, "<div class=\"vwcms_form_test\"><pre>%s</pre></div>",
                  HtmlEscape(sptr));
      return;
   }
   /* the above special case disabled */
   if (*cptr == '!' && strsame (cptr, "!test", -1)) return;

   if (strsame (cptr, "spawned=html", 12) ||
       strsame (cptr, "spawned=plain", 13) ||
       strsame (cptr, "spawned", -1))
   {
      /******************/
      /* spawned output */
      /******************/

      /* special case */
      if (!FormSpawnOutputPtr) return;
      if (strsame (cptr, "spawned=html", 12))
         fputs (FormSpawnOutputPtr, stdout);
      else
         fprintf (stdout,
"<div class=\"vwcms_spawn\" style=\"white-space:pre;%s\">%s</div>",
                     FONT_FAMILY_MONOSPACE,
                     HtmlEscape(FormSpawnOutputPtr));
      return;
   }
   /* the above special case disabled */
   if (*cptr == '!' && strsame (cptr, "!spawn", 6)) return;

   while (*cptr)
   {
      while (!ThisField && *cptr && !isalnum(*cptr))
      {
         if (*cptr == '!')
            FieldDisabled = TRUE;
         else
         if (*cptr == '+')
            FieldRequired = TRUE;
         else
         if (*cptr == '#')
            FieldExample = TRUE;
         cptr++;
      }

      if (strsame (cptr, "action=", 7))
      {
         cptr = TemplateParseParam (cptr+7, FormAction, sizeof(FormAction));
         if (strsame (FormAction, "email=", 6) ||
             strsame (FormAction, "spawn=", 6) ||
             strsame (FormAction, "store=", 6))
            ThisField = FORM_FIELD_ACTION;
         else
         {
            TemplateParamError ("FORM", SubsArg, FormAction);
            return;
         }
      }
      else
      if (strsame (cptr, "align=", 6))
      {
         cptr = TemplateParseParam (cptr+5, FormAlign, sizeof(FormAlign));
         if (tolower(FormAlign[0]) == 'v')
            AlignVert = TRUE;
         else
         if (tolower(FormAlign[0]) == 'h')
            AlignVert = FALSE;
      }
      else
      if (strsame (cptr, "area=", 5))
      {
         ThisField = FORM_FIELD_AREA;
         cptr = TemplateParseParam (cptr+5, FieldName, sizeof(FieldName));
      }
      else
      if (strsame (cptr, "check=", 6))
      {
         ThisField = FORM_FIELD_CHECK;
         cptr = TemplateParseParam (cptr+6, FieldName, sizeof(FieldName));
      }
      else
      if (strsame (cptr, "format=", 7))
      {
         cptr = TemplateParseParam (cptr+7, FormFormat, sizeof(FormFormat));
         if (sptr = GetCgiVarNull ("FORM_TEST_FORMAT"))
         {
            for (aptr = sptr; *aptr && *aptr != '='; aptr++);
            if (*aptr) aptr++; else aptr = sptr;
         }
         else
            aptr = FormFormat;
         if (strsame (aptr, "raw", 3))
            DataFormat = FORM_DATA_FORMAT_RAW;
         else
         if (strsame (aptr, "plain", 5) ||
             strsame (aptr, "text", 4))
            DataFormat = FORM_DATA_FORMAT_PLAIN;
         else
         if (strsame (aptr, "html", 4))
            DataFormat = FORM_DATA_FORMAT_HTML;
         else
         if (strsame (aptr, "xml", 3))
            DataFormat = FORM_DATA_FORMAT_XML;
         else
         if (*aptr)
         {
            TemplateParamError ("FORM", SubsArg, FormFormat);
            return;
         }
      }
      else
      if (strsame (cptr, "goto=", 5))
      {
         cptr = TemplateParseParam (cptr+5, FormGoto, sizeof(FormGoto));
         CgiLibHtmlDeEntify (FormGoto);
      }
      else
      if (strsame (cptr, "insert=", 7))
      {
         /* of main interest when form processing, just validate here */
         ThisField = FORM_FIELD_INSERT;
         cptr = TemplateParseParam (cptr+7, FieldName, sizeof(FieldName));
      }
      else
      if (strsame (cptr, "origin=", 7))
         cptr = TemplateParseSkipParam (cptr+7, NULL);
      else
      if (strsame (cptr, "password=", 9))
      {
         ThisField = FORM_FIELD_PASSWORD;
         cptr = TemplateParseParam (cptr+9, FieldName, sizeof(FieldName));
      }
      else
      if (strsame (cptr, "radio=", 6))
      {
         ThisField = FORM_FIELD_RADIO;
         cptr = TemplateParseParam (cptr+6, FieldName, sizeof(FieldName));
      }
      else
      if (strsame (cptr, "reset[=]", 5))
      {
         ThisField = FORM_FIELD_RESET;
         cptr = TemplateParseParam (cptr+5, FieldName, sizeof(FieldName));
      }
      else
      if (strsame (cptr, "select=", 7))
      {
         ThisField = FORM_FIELD_SELECT;
         cptr = TemplateParseParam (cptr+7, FieldName, sizeof(FieldName));
      }
      else
      if (strsame (cptr, "set[=]", 3))
      {
         ThisField = FORM_FIELD_SET;
         cptr = TemplateParseParam (cptr+3, FieldName, sizeof(FieldName));
      }
      else
      if (strsame (cptr, "size=", 5))
      {
         /* field size or textarea rows-by-cols */
         cptr = TemplateParseRowCol (cptr+5, &TextRows, &TextCols);
         FieldSize = SelectSize = TextRows;
      }
      else
      if (strsame (cptr, "store$", 6) ||
          strsame (cptr, "store", -1))
      {
         if (ContentSectionPtr)
            strcpy (FieldName, ContentSectionPtr->SectionName);
         else
            strcpy (FieldName, ContentSectionName);
         cptr = TemplateParseSkipParam (cptr+5, NULL);
      }
      else
      if (strsame (cptr, "submit=", 7))
      {
         ThisField = FORM_FIELD_SUBMIT;
         cptr = TemplateParseParam (cptr+7, FieldName, sizeof(FieldName));
      }
      else
      if (strsame (cptr, "text=", 5))
      {
         ThisField = FORM_FIELD_TEXT;
         cptr = TemplateParseParam (cptr+5, FieldName, sizeof(FieldName));
      }
      else
      if (strsame (cptr, "value=", 6))
      {
         ValuePtr = cptr + 6;
         break;
      }
      else
      if (*cptr == '=')
      {
         /* synonym for 'value=' */
         ValuePtr = cptr + 1;
         break;
      }
      else
      {
         TemplateParamError ("FORM", SubsArg, cptr);
         return;
      }
   }

   if (FieldSize <= 0 || FieldSize >= 132) FieldSize = 30;
   if (TextRows <= 0 || TextRows >= 24) TextRows = 6;
   if (TextCols <= 0 || TextCols >= 132) TextCols = 60;

   switch (ThisField)
   {
      case FORM_FIELD_ACTION :
         if (FormCount++ > 1)
         {
            /* only one form per section supported */
            TemplateReportError ("FORM", SubsArg,
                                 CONFIG_LANG_PARSE(LangForm,1));
            return;
         }
         if (!strsame (FormAction, "store=", 6) && !FormAction[6])
         {
            TemplateReportError ("FORM", SubsArg,
                                 CONFIG_LANG_PARSE(LangForm,3));
            return;
         }
   }

   switch (ThisField)
   {
      case FORM_FIELD_CHECK :
      case FORM_FIELD_INSERT :
      case FORM_FIELD_PASSWORD :
      case FORM_FIELD_RADIO :
      case FORM_FIELD_SELECT :
      case FORM_FIELD_SUBMIT :
      case FORM_FIELD_TEXT :
         if (!FieldName[0])
         {
            TemplateReportError ("FORM", SubsArg,
                                 CONFIG_LANG_PARSE(LangForm,3));
            return;
         }
   }

   switch (ThisField)
   {
      case FORM_FIELD_CHECK :
      case FORM_FIELD_INSERT :
      case FORM_FIELD_RADIO :
      case FORM_FIELD_SELECT :
         if (!ValuePtr)
         {
            TemplateReportError ("FORM", SubsArg,
                                 CONFIG_LANG_PARSE(LangForm,4));
            return;
         }
   }

   /****************/
   /* form element */
   /****************/

   if (ThisField == FORM_FIELD_ACTION)
   {
      /**********/
      /* action */
      /**********/

      if (FormOpen)
      {
         TemplateAdvise ("E!AZ", CONFIG_LANG_PARSE(LangForm,2));
         return;
      }

      /* see http://code.google.com/apis/recaptcha/docs/customization.html */
      if ((ReCaptchaAfter || ReCaptchaBefore) && VWcmsConfig.ReCaptchaOptions)
         fprintf (stdout,
"<script type=\"text/javascript\">\n\
var RecaptchaOptions = {\n\
    %s\n\
};\n\
</script>\n",
                  VWcmsConfig.ReCaptchaOptions);

      fprintf (stdout,
"<form class=\"vwcms_form\" id=\"vwcms_form\" name=\"vwcms_form\" \
method=\"%s\" action=\"%s?submit\">",
               /* special case - for demonstration purposes */
               FieldExample ? "get" : "post",
               ContentEncodeName("/",ContentSectionName));

      if (sptr = TicketTest(FALSE))
         fprintf (stdout,
"<input type=\"hidden\" name=\"test_ticket\" value=\"%s\" />",
                  sptr); 

      FormDisabled = FieldDisabled;
      FormOpen = TRUE;
      return;
   }

   if (!FormOpen && !FieldExample)
   {
      TemplateAdvise ("E!AZ", CONFIG_LANG_PARSE(LangForm,2));
      return;
   }

   if (ThisField == FORM_FIELD_AREA)
   {
      /************/
      /* textarea */
      /************/

      TemplateParseParam (ValuePtr, FieldValue, sizeof(FieldValue));
      ValuePtr = FieldValue;

      if (RequestMethodPOST && !FieldDisabled)
         if (!(ValuePtr = FormCgiVar (FieldName, 0)))
            if (FieldRequired) FieldAttention = TRUE;

      if (FieldAttention)
         TemplateAdvise ("W!AZ", CONFIG_LANG_PARSE(LangForm,7));

      fprintf (stdout,
"%s<textarea id=\"%s\" name=\"%s\" rows=\"%d\" cols=\"%d\"%s>%s</textarea>%s",
               FieldAttention ? "<span class=\"vwcms_form_attention\">" : "",
               FieldName, FieldName, TextRows, TextCols,
               FieldDisabled ? " disabled=\"disabled\"" : "",
               ValuePtr ? HtmlEscape(ValuePtr) : "",
               FieldAttention ? "</span>" : "");
      return;
   }

   if (ThisField == FORM_FIELD_CHECK)
   {
      /*********/
      /* check */
      /*********/

      if (RequestMethodPOST && !FieldDisabled && FieldRequired)
      {
         /* ensure not only is there a value but is one of those supplied */
         CheckCount = HitCount = 0;
         cptr = ValuePtr;
         while (*cptr)
         {
            cptr = TemplateParseParam (cptr, FieldValue, sizeof(FieldValue));
            CgiLibHtmlDeEntify (FieldValue);
            if (sptr = FormCgiVar (FieldName, ++CheckCount))
            {
               if (FieldValue[0] == '`' && !strcmp (FieldValue+1, sptr))
                  HitCount++;
               else
               if (FieldValue[0] != '`' && !strcmp (FieldValue, sptr))
                  HitCount++;
            }
         }
         if (!HitCount) FieldAttention = TRUE;
      }

      if (FieldAttention)
      {
         TemplateAdvise ("W!AZ", CONFIG_LANG_PARSE(LangForm,7));
         fputs ("<span class=\"vwcms_form_attention\">", stdout);
      }

      CheckCount = 0;
      cptr = ValuePtr;
      while (*cptr)
      {
         CheckCount++;

         cptr = TemplateParseParam (cptr, FieldValue, sizeof(FieldValue));
         CgiLibHtmlDeEntify (FieldValue);
         ValueChecked = ValueDisabled = FALSE;
         for (aptr = FieldValue; *aptr == '!' || *aptr == '`'; aptr++)
         {
            if (*aptr == '!')
               ValueDisabled = TRUE;
            else
            if (*aptr == '`')
               ValueChecked = TRUE;
         }

         if (RequestMethodPOST && !FieldDisabled)
         {
            /* propagate submitted data */
            sptr = FormCgiVar (FieldName, CheckCount);
            if (sptr && !strcmp (aptr, sptr))
               ValueChecked = TRUE;
            else
               ValueChecked = FALSE;
         }

         fprintf (stdout,
"<input type=\"checkbox\" name=\"%s_%d\" value=\"%s\"%s%s />&nbsp;%s",
                  FieldName, CheckCount,
                  HtmlEscape(aptr),
                  ValueChecked ? " checked=\"checked\"" : "",
                  FieldDisabled || ValueDisabled ?
                     " disabled=\"disabled\"" : "",
                  HtmlEscape(aptr));
         if (AlignVert) fputs ("<br />", stdout);
      }

      if (FieldAttention)
      {
         TemplateAdvise ("W!AZ", CONFIG_LANG_PARSE(LangForm,7));
         fputs ("</span>", stdout);
      }
      return;
   }

   if (ThisField == FORM_FIELD_RADIO)
   {
      /*********/
      /* radio */
      /*********/

      sptr = FormOneOf (FieldName, ValuePtr);
      if (RequestMethodPOST && !FieldDisabled && FieldRequired && !sptr)
         FieldAttention = TRUE;

      if (FieldAttention)
      {
         TemplateAdvise ("W!AZ", CONFIG_LANG_PARSE(LangForm,7));
         fputs ("<span class=\"vwcms_form_attention\">", stdout);
      }

      cptr = ValuePtr;
      while (*cptr)
      {
         cptr = TemplateParseParam (cptr, FieldValue, sizeof(FieldValue));
         CgiLibHtmlDeEntify (FieldValue);
         ValueChecked = ValueDisabled = FALSE;
         for (aptr = FieldValue; *aptr == '!' || *aptr == '`'; aptr++)
         {
            if (*aptr == '!')
               ValueDisabled = TRUE;
            else
            if (*aptr == '`')
               ValueChecked = TRUE;
         }

         if (RequestMethodPOST && !FieldDisabled)
         {
            /* propagate submitted data */
            if (sptr && !strcmp (aptr, sptr))
               ValueChecked = TRUE;
            else
               ValueChecked = FALSE;
         }

         fprintf (stdout,
"<input type=\"radio\" name=\"%s\" value=\"%s\"%s%s />&nbsp;%s",
                  FieldName, HtmlEscape(aptr),
                  ValueChecked ? " checked=\"checked\"" : "",
                  FieldDisabled || ValueDisabled ?
                     " disabled=\"disabled\"" : "",
                  HtmlEscape(aptr));
         if (AlignVert) fputs ("<br />", stdout);
      }

      if (FieldAttention)
      {
         TemplateAdvise ("W!AZ", CONFIG_LANG_PARSE(LangForm,7));
         fputs ("</span>", stdout);
      }
      return;
   }

   if (ThisField == FORM_FIELD_SELECT)
   {
      /**********/
      /* select */
      /**********/

      sptr = FormOneOf (FieldName, ValuePtr);
      if (RequestMethodPOST && !FieldDisabled && FieldRequired && !sptr)
         FieldAttention = TRUE;

      if (FieldAttention)
      {
         TemplateAdvise ("W!AZ", CONFIG_LANG_PARSE(LangForm,7));
         fputs ("<span class=\"vwcms_form_attention\">", stdout);
      }

      if (SelectSize)
         fprintf (stdout, "<select name=\"%s\" size=\"%d\"%s>",
                  FieldName, SelectSize,
                  FieldDisabled ? " disabled=\"disabled\"" : "");
      else
         fprintf (stdout, "<select name=\"%s\"%s>",
                  FieldName, FieldDisabled ? " disabled=\"disabled\"" : "");

      cptr = ValuePtr;
      while (*cptr)
      {
         cptr = TemplateParseParam (cptr, FieldValue, sizeof(FieldValue));
         CgiLibHtmlDeEntify (FieldValue);
         ValueChecked = ValueDisabled = FALSE;
         for (aptr = FieldValue; *aptr == '!' || *aptr == '`'; aptr++)
         {
            if (*aptr == '!')
               ValueDisabled = TRUE;
            else
            if (*aptr == '`')
               ValueChecked = TRUE;
         }

         if (RequestMethodPOST && !FieldDisabled)
         {
            /* propagate submitted data */
            if (sptr && !strcmp (aptr, sptr))
               ValueChecked = TRUE;
            else
               ValueChecked = FALSE;
         }

         fprintf (stdout, "<option%s%s>%s</option>",
                  ValueChecked ? " selected=\"selected\"" : "",
                  ValueDisabled ? " disabled=\"disabled\"" : "",
                  HtmlEscape(aptr));
      }

      fputs ("</select>", stdout);

      if (FieldAttention)
      {
         TemplateAdvise ("W!AZ", CONFIG_LANG_PARSE(LangForm,7));
         fputs ("</span>", stdout);
      }
      return;
   }

   if (ThisField == FORM_FIELD_TEXT ||
       ThisField == FORM_FIELD_PASSWORD)
   {
      /********/
      /* text */
      /********/

      TemplateParseParam (ValuePtr, FieldValue, sizeof(FieldValue));
      ValuePtr = FieldValue;

      if (RequestMethodPOST && !FieldDisabled)
         if (!(ValuePtr = FormCgiVar (FieldName, 0)))
            if (FieldRequired) FieldAttention = TRUE;

      if (FieldAttention)
         TemplateAdvise ("W!AZ", CONFIG_LANG_PARSE(LangForm,7));

      fprintf (stdout,
"%s<input type=\"%s\" id=\"%s\" name=\"%s\" size=\"%d\" value=\"%s\"%s />%s",
               FieldAttention ? "<span class=\"vwcms_form_attention\">" : "",
               ThisField == FORM_FIELD_TEXT ? "text" : "password",
               FieldName, FieldName, FieldSize,
               ValuePtr ? HtmlEscape(ValuePtr) : "",
               FieldDisabled ? " disabled=\"disabled\"" : "",
               FieldAttention ? "</span>" : "");
      return;
   }

   if (ThisField == FORM_FIELD_SUBMIT)
   {
      /**********/
      /* submit */
      /**********/

      TemplateParseParam (ValuePtr, FieldValue, sizeof(FieldValue));
      ValuePtr = FieldValue;

      if (FormSubmitDone) FieldDisabled = TRUE;

      if (QandaBefore)
         QandaChallenge ();
      else
      if (ReCaptchaBefore)
         ReCaptchaChallenge ();

      fprintf (stdout,
"<input type=\"submit\" id=\"%s\" name=\"%s\" value=\"%s\"%s%s />",
               FieldName, FieldName, ValuePtr,
               FieldDisabled || FormDisabled ? " disabled=\"disabled\"" : "",
               FieldExample ? " onclick=\"return false\"" : "");

      /* with a form always advise close to the submit button */
      TemplateAdvise ("$$$$<span style=\"font-size:80%\">$</span>");

      if (QandaAfter)
         QandaChallenge ();
      else
      if (ReCaptchaAfter)
         ReCaptchaChallenge ();

      return;
   }

   if (ThisField == FORM_FIELD_RESET)
   {
      /*********/
      /* reset */
      /*********/

      fprintf (stdout,
"<input type=\"reset\" id=\"%s\" name=\"%s\" value=\"%s\"%s />",
               FieldName, FieldName, ValuePtr ? ValuePtr : "",
               FieldDisabled || FormDisabled ? " disabled=\"disabled\"" : "");
      return;
   }

   if (ThisField == FORM_FIELD_SET)
   {
      /************/
      /* fieldset */
      /************/

      if (FieldName[0] && !FieldValue[0])
      {
         strcpy (FieldValue, FieldName);
         FieldName[0] = '\0';
      }

      if (FieldValue[0])
      {
         if (FormFieldSet)
            fputs ("</fieldset>\n", stdout);
         else
            FormFieldSet = TRUE;
         if (FieldName[0])
            fprintf (stdout,
"<fieldset id=\"%s\" name=\"%s\"><legend>%s</legend>\n",
                     FieldName, FieldName, FieldValue);
         else
            fprintf (stdout, "<fieldset><legend>%s</legend>\n", FieldValue);
      }
      else
      {
         fputs ("</fieldset>\n", stdout);
         FormFieldSet = FALSE;
      }
      return;
   }

   if (ThisField == FORM_FIELD_INSERT)
   {
      /**********/
      /* insert */
      /**********/

      if (!ValuePtr) ValuePtr = "";
      if (*ValuePtr != '\"' &&
          !strsame (ValuePtr, "cgi=", 4))
      {
         TemplateParamError ("FORM", SubsArg, ValuePtr);
         return;
      }
   }
}

/*****************************************************************************/
/*
Scan through the entire section text looking for $$FORM$...$$ directives. When
encountered do a sanity check and then if it is a required field ensure it has
some valid data.  Return NULL at the first invalid field.  Return a pointer to
the formatted data if all fields validate and the form is closed correctly.
*/

char* FormProcessPOST ()

{
   BOOL  FieldDisabled = FALSE,
         FieldRequired = FALSE,
         FormError = FALSE,
         FormOpen = FALSE,
         TicketMatch = FALSE;
   int  status,
        CheckCount,
        DataFormat = 0,
        HitCount,
        ThisField = 0;
   char  *aptr, *cptr, *sptr, *tptr,
         *TestTicketPtr = NULL,
         *ValuePtr = NULL;
   char  FieldName [256] = "",
         FieldValue [1024] = "";
   DYSTRING  FormDirect;

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis (FI_LI, "FORM process POST");

   if (!RequestMethodPOST) ErrorExit (SS$_BUGCHECK, FI_LI);

   if (TicketOnPage)
      TestTicketPtr = GetCgiVarNull ("FORM_TEST_TICKET");

   /* look ahead for $$FORM$format=...$$ directive */
   cptr = ContentSectionPtr->TextPtr;

   if (sptr = GetCgiVarNull ("FORM_TEST_FORMAT"))
   {
      for (cptr = sptr; *cptr && *cptr != '='; cptr++);
      if (*cptr) cptr++; else cptr = sptr;
   }
   else
   if (*(cptr = TemplateParseFind (cptr, "$$FORM$format=")))
      cptr += 14;

   if (*cptr)
   {
      if (strsame (cptr, "raw", 3))
         DataFormat = FORM_DATA_FORMAT_RAW;
      else
      if (strsame (cptr, "plain", 5) ||
          strsame (cptr, "text", 4))
         DataFormat = FORM_DATA_FORMAT_PLAIN;
      else
      if (strsame (cptr, "html", 4))
         DataFormat = FORM_DATA_FORMAT_HTML;
      else
      if (strsame (cptr, "xml", 3))
         DataFormat = FORM_DATA_FORMAT_XML;
   }

   if (!DataFormat) DataFormat = FORM_DATA_FORMAT_PLAIN;
   if (WatchEnabled) WatchThis (FI_LI, "FORM format !UL", DataFormat);
   FormFormatData (DataFormat, NULL, NULL, NULL);

   /***********/
   /* process */
   /***********/

   /* using request-persistent storage */
   DYSTRING_VEEMEM_INIT(&FormDirect,0);

   /* process all $$FORM$...$$ directives */
   tptr = ContentSectionPtr->TextPtr;
   while (*(tptr = TemplateParseFind (tptr, "$$FORM$")))
   {
      tptr = TemplateParseDirect (&FormDirect, tptr);

      cptr = DYSTRING_PTR(&FormDirect);
      if (WatchEnabled) WatchThis (FI_LI, "FORM !AZ", cptr);

      /* step over the "FORM$" */
      cptr += 5;

      if (strsame (cptr, "test", -1) ||
          strsame (cptr, "!test", -1) ||
          strsame (cptr, "spawned=", 13) ||
          strsame (cptr, "spawned", -1) ||
          strsame (cptr, "!spawned", -1) ||
          strsame (cptr, "!spawned=", 14))
      {
         /* special cases */
         continue;
      }

      ThisField = 0;
      FieldDisabled = FieldRequired = FALSE;
      FieldName[0] = '\0';
      ValuePtr = NULL;

      while (*cptr)
      {
         /**************/
         /* parameters */
         /**************/

         if (WatchEnabled) WatchThis (FI_LI, "FORM !AZ", cptr);

         while (!ThisField && *cptr && !isalnum(*cptr))
         {
            if (*cptr == '!')
               FieldDisabled = TRUE;
            else
            if (*cptr == '+')
               FieldRequired = TRUE;
            cptr++;
         }

         if (strsame (cptr, "action=", 7))
         {
            ThisField = FORM_FIELD_ACTION;
            cptr = TemplateParseParam (cptr+7, FieldName, sizeof(FieldName));
            FormFormatData (DataFormat, "action", NULL, NULL);
         }
         else
         if (strsame (cptr, "align=", 6))
            cptr = TemplateParseSkipParam (cptr+6, NULL);
         else
         if (strsame (cptr, "area=", 5))
         {
            ThisField = FORM_FIELD_AREA;
            cptr = TemplateParseParam (cptr+5, FieldName, sizeof(FieldName));
            FormFormatData (DataFormat, "area", NULL, NULL);
         }
         else
         if (strsame (cptr, "check=", 6))
         {
            ThisField = FORM_FIELD_CHECK;
            cptr = TemplateParseParam (cptr+6, FieldName, sizeof(FieldName));
         }
         else
         if (strsame (cptr, "format=", 7))
            cptr = TemplateParseSkipParam (cptr+7, NULL);
         else
         if (strsame (cptr, "goto=", 5))
            cptr = TemplateParseSkipParam (cptr+5, NULL);
         else
         if (strsame (cptr, "insert=", 7))
         {
            ThisField = FORM_FIELD_INSERT;
            cptr = TemplateParseParam (cptr+7, FieldName, sizeof(FieldName));
            FormFormatData (DataFormat, "insert", NULL, NULL);
         }
         else
         if (strsame (cptr, "origin=", 7))
            cptr = TemplateParseSkipParam (cptr+7, NULL);
         else
         if (strsame (cptr, "password=", 9))
         {
            ThisField = FORM_FIELD_PASSWORD;
            cptr = TemplateParseParam (cptr+9, FieldName, sizeof(FieldName));
         }
         else
         if (strsame (cptr, "radio=", 6))
         {
            ThisField = FORM_FIELD_RADIO;
            cptr = TemplateParseParam (cptr+6, FieldName, sizeof(FieldName));
            FormFormatData (DataFormat, "radio", NULL, NULL);
         }
         else
         if (strsame (cptr, "reset[=]", 5))
            cptr = TemplateParseSkipParam (cptr+5, NULL);
         else
         if (strsame (cptr, "select=", 7))
         {
            ThisField = FORM_FIELD_SELECT;
            cptr = TemplateParseParam (cptr+7, FieldName, sizeof(FieldName));
            FormFormatData (DataFormat, "select", NULL, NULL);
         }
         else
         if (strsame (cptr, "set[=]", 3))
            cptr = TemplateParseSkipParam (cptr+3, NULL);
         else
         if (strsame (cptr, "size=", 5))
            cptr = TemplateParseSkipParam (cptr+5, NULL);
         else
         if (strsame (cptr, "submit=", 7))
         {
            ThisField = FORM_FIELD_SUBMIT;
            cptr = TemplateParseParam (cptr+7, FieldName, sizeof(FieldName));
            FormFormatData (DataFormat, "submit", NULL, NULL);
         }
         else
         if (strsame (cptr, "text=", 5))
         {
            ThisField = FORM_FIELD_TEXT;
            cptr = TemplateParseParam (cptr+5, FieldName, sizeof(FieldName));
            FormFormatData (DataFormat, "text", NULL, NULL);
         }
         else
         if (strsame (cptr, "value=", 6))
         {
            ValuePtr = (cptr += 6);
            break;
         }
         else
         if (*cptr == '=')
         {
            ValuePtr = (cptr += 1);
            break;
         }
         else
         {
            FormError = TRUE;
            cptr = TemplateParseParam (cptr, NULL, 0);
            continue;
         }
      }

      if (!ThisField) continue;

      if (FieldDisabled)
      {
         FormFormatData (DataFormat, NULL, FieldName, "{disabled}");
         continue;
      }

      /************/
      /* validate */
      /************/

      switch (ThisField)
      {
         case FORM_FIELD_ACTION :

            if (!(strsame (FieldName, "email=", 6) ||
                  strsame (FieldName, "spawn=", 6) ||
                  strsame (FieldName, "store=", 6)))
            {
               FormError = TRUE;
               break;
            }
            if (strsame (FieldName, "store=", 6) && !isalnum(FieldName[6]))
            {
               if (ContentSectionPtr)
                  strcpy (FieldName+6, ContentSectionPtr->SectionName);
               else
                  strcpy (FieldName+6, ContentSectionName);
            }
            FieldName[5] = '\0';
            FormFormatData (DataFormat, NULL, FieldName, FieldName+6);
            FormOpen = TRUE;
            break;

         case FORM_FIELD_AREA :
         case FORM_FIELD_PASSWORD :
         case FORM_FIELD_SUBMIT :
         case FORM_FIELD_TEXT :

            sptr = FormCgiVar (FieldName, 0);

            if (sptr && !FormVetText (sptr))
            {
               FormError = TRUE;
               break;
            }

            /* once hit don't bother looking any further */
            if (TicketOnPage && !TicketMatch && sptr)
            {
               /* there is a $$TICKET$$ in the section */
               if (TestTicketPtr)
               {
                  if (aptr = strstr (sptr, TestTicketPtr))
                  {
                     TicketMatch = TRUE;
                     TicketExcise (sptr, aptr);
                  }
               }
               else
               {
                  status = TicketFile (NULL, sptr);
                  if (VMSok (status)) TicketMatch = TRUE;
               }
            }

            FormFormatData (DataFormat, NULL, FieldName, sptr);
            if (FieldRequired && !sptr) FormError = TRUE;
            break;

         case FORM_FIELD_CHECK :

            CheckCount = HitCount = 0;
            cptr = ValuePtr;
            while (*cptr)
            {
               cptr = TemplateParseParam (cptr, FieldValue, sizeof(FieldValue));
               CgiLibHtmlDeEntify (FieldValue);
               sptr = FormCgiVar (FieldName, ++CheckCount);
               FormFormatData (DataFormat, "check", FieldName, sptr);
               if (!sptr) continue;
               if (FieldValue[0] == '`' && !strcmp (FieldValue+1, sptr))
                  HitCount++;
               else
               if (FieldValue[0] != '`' && !strcmp (FieldValue, sptr))
                  HitCount++;
            }
            if (FieldRequired && !HitCount) FormError = TRUE;
            break;

         case FORM_FIELD_RADIO :
         case FORM_FIELD_SELECT :

            sptr = FormOneOf (FieldName, ValuePtr);
            FormFormatData (DataFormat, NULL, FieldName, sptr);
            if (FieldRequired && !sptr) FormError = TRUE;
            break;

         case FORM_FIELD_INSERT :

            if (*ValuePtr == '\"')
            {
               /* insert a literal string */
               for (sptr = ValuePtr+1; *sptr && *sptr != '\"'; sptr++);
               *sptr = '\0';
               FormFormatData (DataFormat, NULL, FieldName, ValuePtr+1);
            }
            else
            if (strsame (ValuePtr, "cgi=", 4))
            {
               /* insert a CGI value */
               if (!(sptr = GetCgiVarNull (ValuePtr+4))) sptr = "{null}";
               FormFormatData (DataFormat, NULL, FieldName, sptr);
            }
            else
               FormError = TRUE;
            break;

         default:

            ErrorExit (SS$_BUGCHECK, FI_LI);
      }

      if (*cptr) cptr += 2;
   }

   if (FormOpen)
   {
      /******************/
      /* close the form */
      /******************/
 
      FormOpen = FALSE;
      sptr = FormFormatData (DataFormat, NULL, NULL, NULL);

      if (QandaPresent())
      {
         if (cptr = QandaVerify())
         {
            TemplateAdvise ("W!AZ", cptr);
            sleep (REJECT_SLEEP);
            return (NULL);
         }
      }
      else
      if (ReCaptchaPresent())
      {
         if (cptr = ReCaptchaVerify())
         {
            TemplateAdvise ("W!AZ", cptr);
            sleep (REJECT_SLEEP);
            return (NULL);
         }
      }
      else
      if (TicketOnPage)
      {
         if (!TicketMatch)
         {
            TemplateAdvise ("W!AZ", CONFIG_LANG_PARSE(LangModerate,2));
            sleep (REJECT_SLEEP);
            return (NULL);
         }
      }

      if (WatchEnabled)
      {
         WatchThis (FI_LI, "FORM data");
         if (WatchEnabled) WatchDump (sptr, strlen(sptr));;
      }
      if (!FormError) return (sptr);
   }
 
   return (NULL);
}

/*****************************************************************************/
/*
Ensure a text field complies with requirements.
*/

BOOL FormVetText (char *TextPtr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "FormVetText() |%s|\n", TextPtr);

   if (strlen(TextPtr) >> 10 > VWcmsConfig.CommentMaxSizeKbytes)
   {
      TemplateAdvise ("W!AZ (!AZ %X!8XL)",
                      CONFIG_LANG_PARSE(LangForm,6),
                      SysGetMsg(SS$_EXDISKQUOTA), SS$_EXDISKQUOTA);
      sleep (REJECT_SLEEP);
      return (FALSE);
   }

   if (CommentRejectString (TextPtr, CONFIG_LANG_PARSE(LangForm,6)))
   {
      sleep (REJECT_SLEEP);
      return (FALSE);
   }

   if (CommentRejectWord (TextPtr, CONFIG_LANG_PARSE(LangForm,6)))
   {
      sleep (REJECT_SLEEP);
      return (FALSE);
   }

   return (TRUE);
}

/*****************************************************************************/
/*
Construct a CGI FORM variable name from the supplied parameters.  These come
from the POSTed form.  Access the variable and if it doesn't exist or is empty
(all spaces) then return NULL.  If it has content return a pointer to that.
*/

char* FormCgiVar
(
char *FieldName,
int FieldNumber
)
{
   char  *cptr, *sptr;
   char  VarName [32+256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "FormCgiVar()\n");

   if (FieldNumber)
      sprintf (VarName, "FORM_%s_%d", FieldName, FieldNumber);
   else
      sprintf (VarName, "FORM_%s", FieldName);
   sptr = cptr = CgiLibVarNull (VarName);
   if (!sptr) return (NULL);
   while (*sptr && isspace(*sptr)) sptr++;
   if (!*sptr) return (NULL);
   return (cptr);
}

/*****************************************************************************/
/*
Construct a CGI FORM variable as in FormCgiVar().  If the variable has content
compare that to one of the dollar-delimitted strings found in 'ThesePtr'.  If
no match return NULL, otherwise return a pointer to the matched string.
*/

char* FormOneOf
(
char *FieldName,
char *ThesePtr
)
{
   char  *cptr, *sptr;
   char  OneValue [256],
         VarName [32+256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "FormOneOf()\n");

   sprintf (VarName, "FORM_%s", FieldName);
   sptr = CgiLibVarNull (VarName);
   if (!sptr) return (NULL);
   while (*sptr && isspace(*sptr)) sptr++;
   if (!*sptr) return (NULL);
   cptr = ThesePtr;
   while (*cptr)
   {
      cptr = TemplateParseParam (cptr, OneValue, sizeof(OneValue));
      CgiLibHtmlDeEntify (OneValue);
      if (!strcmp (OneValue[0] == '`' ? OneValue+1 : OneValue, sptr))
         return (sptr);
   }
   return (NULL);
}

/*****************************************************************************/
/*
A form has been submitted (POSTed).  Validate it against the form in the
section.  If it validates perform the form action and return true.  If it
doesn't do nothing more and return false.
*/

BOOL FormSubmit (SubmitForm)

{
   BOOL  FormTestIt = FALSE;
   int  FieldAction = 0;
   char  *cptr, *sptr, *tptr,
         *FormPtr;
   char  FormAction [32+256] = "",
         FormGoto [32+256] = "",
         SiteOrigin [256] = "";
   DYSTRING  FormDirect;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "FormSubmit() %d\n", SubmitForm);

   FormSubmitDone = FALSE;

   if (!SubmitForm) return (FALSE);

   if (!RequestMethodPOST) ErrorExit (SS$_BUGCHECK, FI_LI);

   FormPtr = FormProcessPOST ();

   if (!FormPtr) return (FALSE);

   /* using request-persistent storage */
   DYSTRING_VEEMEM_INIT(&FormDirect,0);

   tptr = ContentSectionPtr->TextPtr;
   while (*tptr)
   {
      if (!*(tptr = TemplateParseFind (tptr, "$$FORM$"))) break;

      tptr = TemplateParseDirect (&FormDirect, tptr);

      cptr = DYSTRING_PTR(&FormDirect);
      if (WatchEnabled) WatchThis (FI_LI, "!AZ", cptr);

      /* step over the "FORM$" */
      cptr += 5;

      while (*cptr)
      {
         if (Debug) fprintf (stdout, "|%32.32s|\n", cptr);
         if (strsame (cptr, "action=", 7))
            cptr = TemplateParseParam (cptr+7, FormAction, sizeof(FormAction));
         else
         if (strsame (cptr, "goto=", 5))
         {
            cptr = TemplateParseParam (cptr+5, FormGoto, sizeof(FormGoto));
            CgiLibHtmlDeEntify (FormGoto);
         }
         else
         if (strsame (cptr, "origin=", 7))
            cptr = TemplateParseParam (cptr+7, SiteOrigin, sizeof(SiteOrigin));
         else
         if (strsame (cptr, "test", 4))
         {
            FormTestIt = TRUE;
            cptr = TemplateParseSkipParam (cptr, NULL);
         }
         else
            cptr = TemplateParseSkipParam (cptr, NULL);
      }
   }

   if (!FormAction[0])
   {
      TemplateAdvise ("E!AZ", CONFIG_LANG_PARSE(LangForm,2));
      return (FALSE);
   }

   if (FormTestIt)
   {
      FormSubmitDone = TRUE;
      return (FALSE);
   }

   if (WatchEnabled) WatchThis (FI_LI, "FORM submit !AZ", FormAction);

   if (strsame (FormAction, "email=", 6))
   {
      if (!FormSubmitEmail (FormAction+6, FormPtr, SiteOrigin,
                            VWcmsConfig.PMDFkludge))
         return (FALSE);
   }
   else
   if (strsame (FormAction, "spawn=", 6))
   {
      if (!FormSubmitSpawn (FormAction+6, FormPtr)) return (FALSE);
   }
   else
   if (strsame (FormAction, "store=", 6))
   {
      if (!isalnum(FormAction[6]))
         if (ContentSectionPtr)
            strcpy (FormAction+6, ContentSectionPtr->SectionName);
         else                  
            strcpy (FormAction+6, ContentSectionName);
      if (!FormSubmitStore (FormAction+6, FormPtr)) return (FALSE);
   }
   else
   {
      TemplateAdvise ("E!AZ", CONFIG_LANG_PARSE(LangForm,2));
      return (FALSE);
   }

   FormSubmitDone = TRUE;

   if (!FormGoto[0]) return (FALSE);

   CgiLibResponseRedirect ("%!%s\n%s", FormGoto, CookieHeader());

   return (TRUE);
}

/*****************************************************************************/
/*
Store the form data in a temporary file then spawn the execution of the
specified DCL procedure with that as P1.  The file is deleted after spawn.
Returns true if sucessful.
*/

BOOL FormSubmitSpawn
(
char *FormSpawn,
char *TextPtr
)
{
   int  status;
   char  *aptr, *cptr, *sptr, *zptr,
         *FileNamePtr;
   char  DclCommand [256],
         FileName [256];
   FILE  *FilePtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "FormSubmitSpawn()\n");

   FormSpawnOutputPtr = NULL;

   /* temporary file name */
   sprintf (FileName, "%s%04d%02d%02d%02d%02d%02d%02d.FRM",
            VWcmsConfig.SiteDirectory,
            CurrentVmsNumTime[0], CurrentVmsNumTime[1],
            CurrentVmsNumTime[2], CurrentVmsNumTime[3],
            CurrentVmsNumTime[4], CurrentVmsNumTime[5],
            CurrentVmsNumTime[6]);
   if (Debug) fprintf (stdout, "|%s|\n", FileName);

   FilePtr = fopen (FileName, "w", "ctx=bin", "rfm=stmlf");
   if (!FilePtr)
   {
      status = vaxc$errno;
      TemplateAdvise ("EFORM !AZ (%X!8XL)", SysGetMsg(status), status);
      return (FALSE);
   }

   fputs (TextPtr, FilePtr);
   fclose (FilePtr);

   zptr = (sptr = DclCommand) + sizeof(DclCommand)-1;
   if (sptr < zptr) *sptr++ = '@';
   for (cptr = VWcmsConfig.SiteDirectory;
        *cptr && sptr < zptr;
        *sptr++ = *cptr++);
   if (sptr < zptr) *sptr++ = '_';
   for (cptr = FormSpawn;
        *cptr && *cptr != ' ' && sptr < zptr;
        *sptr++ = *cptr++);
   if (sptr < zptr) *sptr++ = ' ';
   /* P1 is always the data file name */
   for (aptr = FileName; *aptr && sptr < zptr; *sptr++ = *aptr++);
   /* rest (if any) of spawn parameters */
   while (*cptr && *cptr == ' ') cptr++;
   if (*cptr)
   {
      if (sptr < zptr) *sptr++ = ' ';
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   }
   *sptr = '\0';

   if (VMSnok (status = SpawnCommand (DclCommand, &FormSpawnOutputPtr)))
      TemplateAdvise ("EFORM !AZ (%X!8XL)", SysGetMsg(status), status);

   if (Debug) fprintf (stdout, "|%s|\n", FormSpawnOutputPtr);

   delete (FileName);

   return (TRUE);
}

/*****************************************************************************/
/*
Send the form email.  Returns true if sucessful.
*/

BOOL FormSubmitEmail
(
char *FormEmail,
char *TextPtr,
char *SiteOrigin,
char *PMDFkludge
)
{
   int  status;
   char  *cptr, *sptr, *zptr;
   char  MailFrom [256],
         SubjText [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "FormSubmitEmail()\n");

   zptr = (sptr = MailFrom) + sizeof(MailFrom)-1;
   if (PMDFkludge)
   {
      /* for example %IN"", up to and including the first '"' */
      for (cptr = PMDFkludge; *cptr && sptr < zptr; *sptr++ = *cptr++)
         if (cptr > PMDFkludge && *(cptr-1) == '\"') break;
   }
   if (*(cptr = SiteOrigin))
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   else
   if (cptr = VWcmsConfig.SiteOrigin)
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   else
   {
      for (cptr = "VWcms@"; *cptr && sptr < zptr; *sptr++ = *cptr++);
      cptr = GetCgiVarNull ("SERVER_NAME");
      if (!cptr || !*cptr) ErrorExit (SS$_BUGCHECK, FI_LI);
      while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   }
   if (PMDFkludge)
   {
      /* for example %IN"", everything following the first '"' */
      for (cptr = PMDFkludge; *cptr; cptr++)
         if (cptr > PMDFkludge && *(cptr-1) == '\"') break;
       while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   }
   if (sptr >= zptr) ErrorExit (SS$_BUGCHECK, FI_LI);
   *sptr = '\0';

   zptr = (sptr = SubjText) + sizeof(SubjText)-1;
   for (cptr = "VWcms "; *cptr && sptr < zptr; *sptr++ = *cptr++);
   for (cptr = CONFIG_LANG_PARSE(LangForm,1);
        *cptr && sptr < zptr;
        *sptr++ = *cptr++);
   if (sptr < zptr) *sptr++ = ' ';
   for (cptr = CgiRequestUriPtr;
        *cptr && *cptr != '?' && sptr < zptr;
        *sptr++ = *cptr++);
   if (sptr >= zptr) ErrorExit (SS$_BUGCHECK, FI_LI);
   *sptr = '\0';

   cptr = SendMailMessage (MailFrom, FormEmail, NULL, SubjText, TextPtr);

   if (cptr)
      TemplateAdvise ("E!AZ", cptr);
   else
      TemplateAdvise ("I!AZ", CONFIG_LANG_PARSE(LangForm,5));

   /* put it into a cookie so it's available after the redirect */
   TemplateAdvise (NULL);

   return (cptr == NULL);
}

/*****************************************************************************/
/*
Store the form data.  Returns true if sucessful.
*/

BOOL FormSubmitStore
(
char *FormStore,
char *TextPtr
)
{
   int  status;
   char  *cptr, *sptr, *zptr,
         *FileNamePtr;
   FILE  *FilePtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "FormSubmitStore()\n");

   FileNamePtr = MakeFileName (FormStore, ".FRM");

   status = WaitForLock (FileNamePtr, TRUE);
   if (VMSnok (status))
   {
      TemplateAdvise ("EFORM !AZ (%X!8XL)", SysGetMsg(status), status);
      return (FALSE);
   }

   FilePtr = fopen (FileNamePtr, "a", "ctx=bin",
                    "rfm=stmlf", "shr=nil", "fop=cif");
   if (!FilePtr)
   {
      status = vaxc$errno;
      WaitForUnLock ();
      TemplateAdvise ("EFORM !AZ (%X!8XL)", SysGetMsg(status), status);
      return (FALSE);
   }

   fputs (TextPtr, FilePtr);
   fputs ("\n", FilePtr);

   fclose (FilePtr);

   if (chmod (FileNamePtr, 0700)) ErrorExit (vaxc$errno, FI_LI);

   WaitForUnLock ();

   TemplateAdvise ("I!AZ", CONFIG_LANG_PARSE(LangForm,5));

   return (TRUE);
}

/*****************************************************************************/
/*
Build a null-terminated string in dynamic memory, containing the POSTed form
data in the requested format.
*/

char* FormFormatData
(
int DataFormat,
char *FieldType,
char *FieldName,
char *FieldValue
)
{
   static BOOL  DataOpen = FALSE;
   static char  *XmlFieldTypePtr;
   static DYSTRING  FormData;

   char  *cptr, *sptr,
         *HostNamePtr,
         *PathPtr,
         *RemoteHostPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "FormFormatData()\n");

   if (!DataFormat) return (DYSTRING_PTR(&FormData));

   if (!DataOpen)
   {
      /********/
      /* open */
      /********/

      HostNamePtr = RemoteHostPtr = GetCgiVar ("REMOTE_HOST");
      if (!HostNamePtr) ErrorExit (SS$_BUGCHECK, FI_LI);
      for (cptr = HostNamePtr;
           *cptr && (*cptr == '.' || isdigit(*cptr));
           cptr++);
      if (!*cptr) cptr = HostNameByAddr (HostNamePtr);
      if (cptr) HostNamePtr = cptr; 

      PathPtr = cptr  = GetCgiVar ("REQUEST_URI");
      while (*cptr && *cptr != '?') cptr++;
      *cptr = '\0';

      /* using request-persistent storage */
      DYSTRING_VEEMEM_INIT(&FormData,0);
   }

   if (DataFormat == FORM_DATA_FORMAT_RAW)
   {
      /*******/
      /* raw */
      /*******/

      if (!FieldType && !FieldName && !FieldValue)
      {
         DataOpen = FALSE;
         return (DYSTRING_PTR(&FormData));
      }

      if (!DataOpen)
      {
         /* only need to do it the once! */
         DataOpen = TRUE;
         sptr = BuildString (&FormData, PostBufferPtr);
      }
      else
         sptr = BuildString (&FormData, "");
      return (sptr);
   }

   if (DataFormat == FORM_DATA_FORMAT_PLAIN)
   {
      /*********/
      /* plain */
      /*********/

      if (!FieldType && !FieldName && !FieldValue)
      {
         DataOpen = FALSE;
         return (DYSTRING_PTR(&FormData));
      }

      if (!DataOpen)
      {
         DataOpen = TRUE;
         BuildString (&FormData, "Time: ", TimeString(0,0), "\n");
         BuildString (&FormData, "Server: //", GetCgiVar ("SERVER_NAME"),
                      PathPtr, " (", SoftwareId, ")\n");
         if (HostNamePtr == RemoteHostPtr)
            BuildString (&FormData, "Client: ", HostNamePtr, "\n");
         else
            BuildString (&FormData, "Client: ", HostNamePtr, " (", RemoteHostPtr, ")\n");
         BuildString (&FormData, "Agent: ", GetCgiVar ("HTTP_USER_AGENT"), "\n");
      }

      if (FieldType) sptr = BuildString (&FormData, FieldType, "/");
      if (FieldName)
         sptr = BuildString (&FormData, FieldName, ": ", FieldValue, "\n");
      else
      if (FieldValue)
         sptr = BuildString (&FormData, FieldValue, "\n");

      return (sptr);
   }

   if (DataFormat == FORM_DATA_FORMAT_HTML)
   {
      /********/
      /* html */
      /********/

      if (!FieldType && !FieldName && !FieldValue)
      {
         if (DataOpen) BuildString (&FormData, "</table>\n");
         DataOpen = FALSE;
         return (DYSTRING_PTR(&FormData));
      }

      if (!DataOpen)
      {
         DataOpen = TRUE;
         BuildString (&FormData, 
"<table cellpadding=\"0\" cellspacing=\"5\" border=\"0\">\n");
         BuildString (&FormData, 
"<tr><td></td><th align=\"right\">Time:</th><td>",
                      TimeString(0,0), "</td></tr>\n");
         BuildString (&FormData, 
"<tr><td></td><th align=\"right\">Server:</th><td>//",
                      GetCgiVar ("SERVER_NAME"),
                      PathPtr, " (", SoftwareId, ")</td></tr>\n");
         BuildString (&FormData, 
"<tr><td></td><th align=\"right\">Client:</th><td>");
         if (HostNamePtr == RemoteHostPtr)
            BuildString (&FormData, HostNamePtr, "</td></tr>\n");
         else
            BuildString (&FormData, HostNamePtr, " (", RemoteHostPtr, ")</td></tr>\n");
         BuildString (&FormData, 
"<tr><td></td><th align=\"right\">Agent:</th><td>",
                      HtmlEscape(GetCgiVar("HTTP_USER_AGENT")),
                      "</td></tr>\n");
      }

      if (FieldType)
         sptr = BuildString (&FormData, "<tr><td>", FieldType, "</td>");
      if (FieldName && FieldValue)
         sptr = BuildString (&FormData, 
"<th align=\"right\">", HtmlEscape(FieldName), ":</th>",
                             strchr(FieldValue,'\n') ?
                                "<td style=\"white-space:pre;\">" : "<td>",
                             HtmlEscape(FieldValue),
                             "</td></tr>\n");
      else
      if (FieldName)
         sptr = BuildString (&FormData,
"<th align=\"right\">", HtmlEscape(FieldName), ":</th><td>",
                             NULL, "</td></tr>\n");
      else
      if (FieldValue)
         sptr = BuildString (&FormData, "<th></th><td>",
                             strchr(FieldValue,'\n') ? "<pre>" : "",
                             FieldValue ? HtmlEscape(FieldValue) : NULL,
                             strchr(FieldValue,'\n') ? "</pre>" : "",
                             "</td></tr>\n");

      return (sptr);
   }

   if (DataFormat == FORM_DATA_FORMAT_XML)
   {
      /*******/
      /* XML */
      /*******/

      if (!FieldType && !FieldName && !FieldValue)
      {
         if (DataOpen)
         {
            if (XmlFieldTypePtr) BuildString (&FormData, "</", XmlFieldTypePtr, ">\n");
            BuildString (&FormData, "</data>\n");
         }
         DataOpen = FALSE;
         XmlFieldTypePtr = NULL;
         return (DYSTRING_PTR(&FormData));
      }

      if (!DataOpen)
      {
         DataOpen = TRUE;
         XmlFieldTypePtr = NULL;
         BuildString (&FormData, "<?xml version=\"1.0\" encoding=\"",
                      VWcmsConfig.LangCharSet, "\"?>\n<data>\n");
         BuildString (&FormData, "<time>", TimeString(0,0), "</time>\n");
         BuildString (&FormData, "<server>//",
                      GetCgiVar ("SERVER_NAME"),
                      PathPtr, " (", SoftwareId,
                      ")</server>\n");
         BuildString (&FormData, "<client>");
         if (HostNamePtr == RemoteHostPtr)
            BuildString (&FormData, HostNamePtr, "</client>\n");
         else
            BuildString (&FormData, HostNamePtr, " (", RemoteHostPtr, ")</client>\n");
         BuildString (&FormData, "<agent>",
                      HtmlEscape(GetCgiVar("HTTP_USER_AGENT")),
                      "</agent>\n");
      }

      if (FieldType)
      {
         if (XmlFieldTypePtr) BuildString (&FormData, "</", XmlFieldTypePtr, ">\n");
         XmlFieldTypePtr = FieldType;
         sptr = BuildString (&FormData, "<", FieldType, ">");
      }
      if (FieldName)
         sptr = BuildString (&FormData, "<name>", HtmlEscape(FieldName), "</name><value>",
                             FieldValue ? HtmlEscape(FieldValue) : NULL,
                             "</value>");
      else
      if (FieldValue)
         sptr = BuildString (&FormData, "<value>", HtmlEscape(FieldValue), "</value>");

      return (sptr);
   }

   ErrorExit (SS$_BUGCHECK, FI_LI);
}

/*****************************************************************************/

