/*****************************************************************************/
/*
                                  other.c

Miscellaneous functions for VWCMS.


COPYRIGHT
---------
Copyright (C) 2008-2011 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it under the
conditions of the GNU GENERAL PUBLIC LICENSE, version 3, or any later version.
http://www.gnu.org/licenses/gpl.txt


VERSION HISTORY
---------------
29-AUG-2011  MGD  bugfix; OutputFromTo() leading '/' processing (on root site)
21-MAY-2011  MGD  OutputFromTo() now two successive '//' means don't touch!
23-APR-2010  MGD  OutputFromTo() two successive '_' means don't touch!
03-APR-2010  MGD  OutputFromTo() to output content and
                    dynamically maps the site name into "SRC=" and "HREF="
                  IsAnImage() dynamically maps the site name into "SRC="
24-MAR-2008  MGD  initial
*/

/*****************************************************************************/

#ifdef VWCMS_VMS_V62
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60200000
#undef __CRTL_VER
#define __CRTL_VER 60200000
#endif

#ifndef __VAX
#   pragma nomember_alignment
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stat.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <types.h>

/* Internet-related header files */
#include <socket.h>
#include <in.h>
#include <inet.h>

/**
HP C V7.1-015 on OpenVMS Alpha V8.3 reports the following (apparently) because
of a nesting error in RESOLV.H due to the __CRTL_VER constraint that WASD
applications are built with.  This does not happen with Compaq C V6.4-005 on
OpenVMS Alpha V7.3-2 for instance.  Just avoid the error by pretending the
header file is already loaded.  Doesn't *seem* to cause side-effects.

#pragma __member_alignment __restore
........^
%CC-W-ALIGNPOP, This "restore" has underflowed the member alignment's stack. No corresponding "save" was found.
at line number 451 in module RESOLV of text library SYS$COMMON:[SYSLIB]DECC$RTLD
**/

#define __RESOLV_LOADED 1
#include <netdb.h>

/* VMS related header files */
#include <descrip.h>
#include <iodef.h>
#include <lckdef.h>
#include <lkidef.h>
#include <lnmdef.h>
#include <lib$routines.h>
#include <mailmsgdef.h>
#include <fab.h>
#include <prvdef.h>
#include <rab.h>
#include <rms.h>
#include <ssdef.h>
#include <starlet.h>
#include <stsdef.h>

/* application header file */
#include "vwcms.h"
#include "config.h"
#include "content.h"
#include "cookie.h"
#include "other.h"
#include "cgilib.h"

#define FI_LI "OTHER", __LINE__

/* global storage */

char Base64Chars[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ\
abcdefghijklmnopqrstuvwxyz0123456789+/";

char Base64UrlSafeChars[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ\
abcdefghijklmnopqrstuvwxyz0123456789-_";

#define XX 255

static unsigned char  Base64Table [] =
{
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
   XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,62, XX,XX,XX,63,
   52,53,54,55, 56,57,58,59, 60,61,XX,XX, XX,XX,XX,XX,
   XX, 0, 1, 2,  3, 4, 5, 6,  7, 8, 9,10, 11,12,13,14,
   15,16,17,18, 19,20,21,22, 23,24,25,XX, XX,XX,XX,XX,
   XX,26,27,28, 29,30,31,32, 33,34,35,36, 37,38,39,40,
   41,42,43,44, 45,46,47,48, 49,50,51,XX, XX,XX,XX,XX 
};

unsigned long  SysLckMask [2] = { PRV$M_SYSLCK, 0 };

#undef lksb
struct lksb
{
   unsigned short int lksb$w_status;
   unsigned short int lksb$w_reserved;
   unsigned int lksb$l_lkid;
   unsigned char lksb$b_valblk [16];
};

struct lksb  WaitForLockSb;

/* prototypes */

/* external storage */

extern BOOL  Debug,
             WatchEnabled;

extern int  VmsVersion;

extern unsigned long  CurrentSecondsTime;

extern char  *CgiEnvironmentPtr,
             *CgiRequestUriPtr,
             *LastModifiedNow,
             *VwcmsScriptPath,
             *VwcmsStatusPtr;

extern char  ContentSiteName[];

extern char  *LangArray[];

extern char  SoftwareCopy[],
             SoftwareEnv[],
             Utility[];

extern CONFIG_DATA  VWcmsConfig;

/*****************************************************************************/
/*
This is the fundamental output function for VWcms content.  Although VWcms does
fputs() and fprintf() directly to <stdout> in lots of places this function
outputs and massages the majority of VWcms site content.

The 'fptr' (from pointer) is the first character to be output and the 'tptr'
(to pointer) is the next character along from the last to be output.

To allow VWcms to dynamically map the site name "/_..." into "SRC=" and "HREF="
it is necessary to have it massage those attributes.  It expunges anything
between the "/_" and the following '/' and substitutes the current site name
(in the same way as does IsAnImage()).
*/

void OutputFromTo
(
char *fptr,
char *tptr
)

{
   char  ch;
   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   cptr = fptr;
   while (cptr < tptr)
   {
      /* search for start-of-tag */
      while (cptr < tptr && *cptr != '<') cptr++;
      if (cptr >= tptr) break;

      /* search for end-of-tag while checking for "src=" and "href=" */
      while (cptr < tptr && *cptr != '>' &&
            *(ULONGPTR)cptr != 'src=' && *(ULONGPTR)cptr != 'SRC=' &&
            *(ULONGPTR)cptr != 'href' && *(ULONGPTR)cptr != 'HREF')
         cptr++;
      if (cptr >= tptr) break;

      /* if reached end-of-tag before hitting attribute */
      if (*cptr == '>') continue;

      /* must have hit a "src=" or "href", drop thru */
      if (WatchEnabled) WatchThis (FI_LI, "URI |!16AZ", cptr);

      if (*(ULONGPTR)cptr == 'src=' || *(ULONGPTR)cptr == 'SRC=')
         cptr += 4;
      else
      if (*(ULONGPTR)(cptr+1) == 'ref=' || *(ULONGPTR)(cptr+1) == 'REF=')
         cptr += 5;
      else
      {
         /* remotely possible if the "href" turned out not to be "href=" */
         cptr++;
         continue;
      }
      if (*cptr == '\"' || *cptr == '\'') cptr++;

      for (sptr = cptr; *sptr && *sptr != ':' && *sptr != '/'; sptr++);
      /* if it looks like a full URL */
      if (*sptr == ':') continue;

      /* output up to here */
      ch = *cptr;
      *cptr = '\0';
      fputs (fptr, stdout);
      *cptr = ch;

      /* ensure the current site name is used */
      if (*(USHORTPTR)cptr == '//')
      {
         /* do not touch this local resource, skip one slash */
         if (cptr < tptr) cptr++;
      }
      else
      if (*(USHORTPTR)cptr == '/_')
      {
         /* explicit reference to a resource */
         for (cptr += 2; cptr < tptr && *cptr != '/'; cptr++);
         if (*cptr && cptr < tptr) cptr++;
         fputs (ContentEncodeName("/_",NULL), stdout);
      }
      else
      if (*cptr == '/')
      {
         /* explicit reference to a section */
         if (VWcmsConfig.SiteRoot)
            cptr++;
         else
         {
            for (cptr++; cptr < tptr && *cptr != '/'; cptr++);
            if (*cptr && cptr < tptr) cptr++;
         }
         fputs (ContentEncodeName("/",NULL), stdout);
      }
      else
      {
         /* establish whether it looks more like a resource or a section */
         for (sptr = cptr; *sptr && sptr < tptr; sptr++)
            if (*sptr == '\"' || *sptr == '\'' ||
                *sptr == ' ' || *sptr == '\t' ||
                *sptr == '\n') break;
         if (sptr < tptr)
            while (sptr > cptr && *sptr != '/' && *sptr != '.') sptr--;

         /* if looks like a file then it's a resource */
         if (sptr > cptr && sptr < tptr && *sptr == '.')
            fputs (ContentEncodeName("/_",NULL), stdout);
         else
            fputs (ContentEncodeName("/",NULL), stdout);

         if (*(USHORTPTR)cptr == './')
         {
            if (cptr < tptr) cptr++;
            if (cptr < tptr) cptr++;
         }
      }

      fptr = cptr;
   }

   if (fptr < tptr)
   {
      /* output remainder */
      ch = *tptr;
      *tptr = '\0';
      fputs (fptr, stdout);
      *tptr = ch;
   }
}

/*****************************************************************************/
/*
When a 'button' is to be rendered as an image return a pointer to all the
supplied attributes in a dynamic, null-terminated string.  If not an image then
return NULL.

To allow VWcms to dynamically map the site name "/_..." it is necessary to have
it massage the "SRC=" of the image.  It expunges anything between the "/_" and
the following '/' and substritues the current site name (in the way same as
does OutputFromTo()).
*/

char* IsAnImage (char *StringPtr)

{
   int  len;
   char  *aptr, *cptr, *sptr, *zptr,
         *ImgPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "IsAnImage()\n");

   cptr = StringPtr;

   if (*cptr == '&' && strsame (cptr, "&lt;img", 7))
   {
      /* local copy (we'll need it later anyway) */
      sptr = CgiLibVeeMemCalloc (strlen(cptr));
      if (!sptr) ErrorExit (vaxc$errno, FI_LI);
      strcpy (sptr, cptr);
      CgiLibHtmlDeEntify (cptr = sptr);
   }
   else
   if (*cptr != '<' || !strsame (cptr, "<img", 4))
      return (NULL);

   if (Debug) fprintf (stdout, "|%s|\n", cptr);

   len = strlen (cptr);
   ImgPtr = sptr = CgiLibVeeMemCalloc (len + 128);
   if (!sptr) ErrorExit (vaxc$errno, FI_LI);
   zptr = sptr + len + 127;

   /* trim any leading <img../> tag */
   cptr += 4;
   while (*cptr && *cptr == ' ') cptr++;
   while (*cptr && sptr < zptr &&
          *(ULONGPTR)cptr != 'src=' &&
          *(ULONGPTR)cptr != 'SRC=')
      *sptr++ = *cptr++;
   if (*cptr && sptr < zptr) *sptr++ = *cptr++;
   if (*cptr && sptr < zptr) *sptr++ = *cptr++;
   if (*cptr && sptr < zptr) *sptr++ = *cptr++;
   if (*cptr && sptr < zptr) *sptr++ = *cptr++;
   if ((*cptr == '\"' || *cptr == '\'') && sptr < zptr) *sptr++ = *cptr++;
   if (*(USHORTPTR)cptr == '/_')
   {
      if (sptr < zptr) *sptr++ = *cptr++;
      if (sptr < zptr) *sptr++ = *cptr++;
      while (*cptr && *cptr != '/') cptr++;
      for (aptr = ContentSiteName; *aptr && sptr < zptr; *sptr++ = *aptr++);
   }
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   if (sptr >= zptr) return (NULL);
   /* trim any trailing <img../> tag */
   while (sptr > ImgPtr && *sptr != '>') sptr--;
   if (sptr > ImgPtr) sptr--;
   if (sptr > ImgPtr && *sptr == '/') sptr--;
   while (sptr > ImgPtr && *sptr == ' ') sptr--;
   if (sptr > ImgPtr) sptr++;
   *sptr = '\0';

   if (Debug) fprintf (stdout, "|%s|\n", ImgPtr);
   return (ImgPtr); 
}

/*****************************************************************************/
/*
Implement a redirect.  '/' a local URI, '!' the following URL plus append the
original request URI, anything else an external URL.
*/

void Unknown302 ()

{
   /*********/
   /* begin */
   /*********/

   /* section not found and redirection configured */
   CgiLibResponseRedirect (*VWcmsConfig.Unknown302 == '/' ?
                              "%!%s%s\n%s" : "%s%s\n%s",
                           *VWcmsConfig.Unknown302 == '!' ?
                              VWcmsConfig.Unknown302+1 : "",
                           *VWcmsConfig.Unknown302 == '!' ?
                              CgiRequestUriPtr : VWcmsConfig.Unknown302,
                           CookieHeader());
}

/*****************************************************************************/
/*
Process a VMS-style date/time string in the format "31-MAR-2005 20:35:16.32"
returning a pointer to an allocated string in the RFC/Unix-style format
"31 Mar 2005 20:35:16".  If 'VmsDateString' is NULL then use current time.
*/

char* RfcVmsDate (char *VmsDateString)

{
   static char  VmsDate [24];
   static $DESCRIPTOR (VmsDateDsc, VmsDate);
   static $DESCRIPTOR (VmsDateTimeFaoDsc, "!%D\0");

   int  cnt, dd, yyyy, hh, mm, ss;
   char  *cptr, *sptr;
   char  mon [32],
         RfcDateString [64];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "RfcVmsDate() |%s|\n", VmsDateString);

   if (!VmsDateString)
   {
      sys$fao (&VmsDateTimeFaoDsc, 0, &VmsDateDsc, 0);
      VmsDateString = VmsDate;
   }

   cnt = sscanf (VmsDateString, "%2d-%3s-%4d %2d:%2d:%2d",
                 &dd, &mon, &yyyy, &hh, &mm, &ss);
   if (cnt != 6) cnt = 0;

   if (dd < 1 || dd > 31) cnt = 0;
   if (strlen(mon) != 3) cnt = 0;
   if (yyyy < 1970 || yyyy > 2099) cnt = 0;
   if (hh < 0 || hh > 24) cnt = 0;
   if (mm < 0 || mm > 59) cnt = 0;
   if (ss < 0 || ss > 59) cnt = 0;

   if (cnt)
   {
      mon[0] = toupper(mon[0]);
      mon[1] = tolower(mon[1]);
      mon[2] = tolower(mon[2]);

      cnt = sprintf (cptr = RfcDateString, "%02d %s %d %02d:%02d:%02d",
                     dd, mon, yyyy, hh, mm, ss);
   }
   else
      cnt = strlen(cptr = VmsDateString);

   sptr = CgiLibVeeMemCalloc (cnt);
   if (!sptr) ErrorExit (vaxc$errno, FI_LI);
   memcpy (sptr, cptr, cnt);

   if (Debug) fprintf (stdout, "|%s|\n", sptr);

   return (sptr);
}

/*****************************************************************************/
/*
Watch CGI values as they are accessed.
*/

char* GetCgiVar (char *NamePtr)

{
   char  *ValuePtr;

   /*********/
   /* begin */
   /*********/

   ValuePtr = CgiLibVeeMemVar (NamePtr);
   if (WatchEnabled) WatchThis (FI_LI, "CGIVAR !AZ=!AZ", NamePtr, ValuePtr);
   return (ValuePtr);
}

/*****************************************************************************/
/*
Watch CGI values as they are accessed.
*/

char* GetCgiVarNull (char *NamePtr)

{
   char  *ValuePtr;

   /*********/
   /* begin */
   /*********/

   ValuePtr = CgiLibVeeMemVarNull (NamePtr);
   if (WatchEnabled)
      WatchThis (FI_LI, "CGIVAR !AZ=!AZ",
                 NamePtr, ValuePtr ? ValuePtr : "(null)");
   return (ValuePtr);
}

/*****************************************************************************/
/*
Return a pointer to a VeeMem string containing the formatted time.  If the
hours and minutes are both zero then just provide the date (no hh:mm).
*/

char* TimeString
(
char *FormatString,
unsigned long SecondsTime
)
{
   int  len = 0;
   char  *cptr;
   char  TimeBuffer [256];
   struct tm  *tm_ptr; 

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TimeString()\n");

   if (!SecondsTime) SecondsTime = CurrentSecondsTime;
   tm_ptr = localtime (&SecondsTime);

   cptr = FormatString;
   if (!cptr || !*cptr)
   {
      if (tm_ptr->tm_hour || tm_ptr->tm_min)
         cptr = VWcmsConfig.LangTimeFormat;
      else
         cptr = VWcmsConfig.LangDateFormat;
   }
   else
   if (strsame (cptr, "VMS", 3) && (!*(cptr+3) || isdigit(*(cptr+3))))
   {
      if (isdigit(*(cptr+3)))
      {
         len = atoi(cptr+3);
         if (len <= 0 || len > 23) len = 17;
      }
      else
         len = 17;
      if (tm_ptr->tm_hour || tm_ptr->tm_min)
         cptr = "%.2d-%b-%Y %H:%M:%S.00";
      else
         cptr = "%.2d-%b-%Y";
   }      
   else
   if (strsame (cptr, "%VMS", 4) && (!*(cptr+4) || isdigit(*(cptr+4))))
   {
      if (isdigit(*(cptr+4)))
      {
         len = atoi(cptr+4);
         if (len <= 0 || len > 23) len = 17;
      }
      else
         len = 17;
      if (tm_ptr->tm_hour || tm_ptr->tm_min)
         cptr = "%.2d-%b-%Y %H:%M:%S.00";
      else
         cptr = "%.2d-%b-%Y";
   }      

   if (!strftime (TimeBuffer, sizeof(TimeBuffer), cptr, tm_ptr))
      sprintf (TimeBuffer, "!strftime() \'%s\' %s", cptr, strerror(errno));
   else
   if (len)
   {
      if (TimeBuffer[0] == '0') TimeBuffer[0] = ' ';
      TimeBuffer[3] = toupper(TimeBuffer[3]);
      TimeBuffer[4] = toupper(TimeBuffer[4]);
      TimeBuffer[5] = toupper(TimeBuffer[5]);
      TimeBuffer[len] = '\0';
   }

   cptr = CgiLibVeeMemCalloc (strlen(TimeBuffer));
   if (!cptr) ErrorExit (SS$_BUGCHECK, FI_LI);
   strcpy (cptr, TimeBuffer);

   return (cptr);
}

/*****************************************************************************/
/*
Using strptime() get a time from a string (just wraps strptime()).
*/

/* VMS V6.2 was the first to provide strptime() so we'll provide 'our' own */
char* GNU_strptime (char*, char*, struct tm*);

char* GetTimeString
(
char *TimeString,
struct tm *tm_ptr
)
{
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "GetTimeString()\n");

   memset (tm_ptr, 0, sizeof(struct tm));

   /* allow for a leading weekday */
#if __CRTL_VER >= 60200000
   if (isalpha(*TimeString))
      cptr = strptime (TimeString, "%A, %d %B %Y %H:%M", tm_ptr);
   else
      cptr = strptime (TimeString, "%d %B %Y %H:%M", tm_ptr);
   if (!cptr)
   {
      /* just a date? */
      if (isalpha(*TimeString))
         cptr = strptime (TimeString, "%A, %d %B %Y", tm_ptr);
      else
         cptr = strptime (TimeString, "%d %B %Y", tm_ptr);
   }
#else
   if (isalpha(*TimeString))
      cptr = GNU_strptime (TimeString, "%A, %d %B %Y %H:%M", tm_ptr);
   else
      cptr = GNU_strptime (TimeString, "%d %B %Y %H:%M", tm_ptr);
   if (!cptr)
   {
      /* just a date? */
      if (isalpha(*TimeString))
         cptr = GNU_strptime (TimeString, "%A, %d %B %Y", tm_ptr);
      else
         cptr = GNU_strptime (TimeString, "%d %B %Y", tm_ptr);
   }
#endif

   return (cptr);
}

/*****************************************************************************/
/*
Return a dynamic string containing the size in MBytes, kBytes, etc.
*/

char* SizeBytes (SizeBytes)

{
   float  fbytes;
   char  SizeString [32];
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SizeBytes()\n");

   if (SizeBytes > 1048575)
   {
      if ((fbytes = (float)(SizeBytes) / 1048576.0) >= 100.0)
         sprintf (SizeString, "%.0f MBytes", fbytes);
      else
         sprintf (SizeString, "%.1f MBytes", fbytes);
   }
   else
   if (SizeBytes > 1023)
   {
      if ((fbytes = (float)(SizeBytes) / 1024.0) >= 100.0)
         sprintf (SizeString, "%.0f kBytes", fbytes);
      else
         sprintf (SizeString, "%.1f kBytes", fbytes);
   }
   else
      sprintf (SizeString, "%d Bytes", SizeBytes);

   cptr = CgiLibVeeMemCalloc(strlen(SizeString));
   if (!cptr) ErrorExit (vaxc$errno, FI_LI);
   strcpy (cptr, SizeString);

   return (cptr);
}

/*****************************************************************************/
/*
Generate a file name from the supplied name (usually the section name).  If
required makes it ODS-2 compliant and if naturally longer than 39 characters
compresses its size by removing vowels.  The file type (e.g. ".LIS", ".CMT") is
supplied so that compliant file names can be generated for facilities (i.e.
currently TemplateHit()) other than the reader comment.  Name of file is
prefixed by an underscore so it cannot be used as a resource.
*/

char* MakeFileName
(
char *SectionName,
char *FileType
)
{
   int  NameLength;
   char  ch;
   char  *cptr, *sptr, *zptr,
         *NamePtr;
   char  FileName [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MakeFileName() |%s|\n", SectionName);

   zptr = (sptr = FileName) + sizeof(FileName)-1;
   for (cptr = VWcmsConfig.SiteDirectory;
        *cptr && sptr < zptr;
        *sptr++ = *cptr++);
   NamePtr = sptr;
   /* prefix with an underscore */
   if (sptr < zptr) *sptr++ = '_';
   /* the leading slash is always a given */
   if (*(cptr = SectionName) == '/') cptr++;
   if (VWcmsConfig.OnDiskStruct == 5)
   {
      while (*cptr && sptr < zptr)
      {
         ch = *cptr++;
         if (isalnum(ch))
            *sptr++ = ch;
         else
         if (ch == '-' || ch == '_')
            *sptr++ = ch;
         else
         if (ch == ' ')
            *sptr++ = '_';
         else
         if (ch == '/')
            *sptr++ = '$';
      }
   }
   else
   {
      for (NameLength = 1; *cptr; cptr++)
         if (isalnum(*cptr) || *cptr == '-' || *cptr == '_'  ||
                               *cptr == ' ' || *cptr == '/') NameLength++;
      if (*(cptr = SectionName) == '/') cptr++;
      while (*cptr && sptr < zptr && (sptr - NamePtr < 39))
      {
         ch = toupper(*cptr++);
         if (isalnum(ch))
         {
            if (NameLength <= 39)
               *sptr++ = ch;
            else
            if (ch != 'A' && ch != 'E' && ch != 'I' && ch != 'O' && ch != 'U')
               *sptr++ = ch;
         }
         else
         if (ch == '-' || ch == '_')
            *sptr++ = ch;
         else
         if (ch == ' ')
            *sptr++ = '_';
         else
         if (ch == '/')
            *sptr++ = '$';
      }
   }
   for (cptr = FileType; *cptr && sptr < zptr; *sptr++ = *cptr++);
   if (sptr >= zptr) ErrorExit (SS$_BUGCHECK, FI_LI);
   *sptr = '\0';

   cptr = CgiLibVeeMemCalloc(sptr-FileName+1);
   strcpy (cptr, FileName);
   if (!cptr) ErrorExit (vaxc$errno, FI_LI);

   if (WatchEnabled)
      WatchThis (FI_LI, "FILENAME \"!AZ\" !AZ", SectionName, cptr);

   return (cptr);
}

/*****************************************************************************/
/*
Build a null-terminated string in dynamic memory.  A variable number of
null-terminated string parameters can be supplied with each being concatenated. 
If a NULL first parameter then reset the string build.  Always returns a
pointer to the current (or previous) string (even if that's NULL).
*/

char* BuildString
(
DYSTRING *dyptr,
char *TextPtr,
...
)
{
   int  argcnt;
   char  *cptr, *sptr, *zptr;
   va_list  argptr;

   /*********/
   /* begin */
   /*********/

   va_count (argcnt);

   if (Debug) fprintf (stdout, "BuildString() %d\n", argcnt);

   if (!dyptr->csize) dyptr->csize = 1024;

   if (!dyptr->bsize)
   {
      /* initial allocation (in case of empty string) */
      dyptr->bsize += dyptr->csize;
      if (dyptr->vmem)
         dyptr->bptr = CgiLibVeeMemRealloc (dyptr->bptr, dyptr->bsize+32);
      else
         dyptr->bptr = realloc (dyptr->bptr, dyptr->bsize+32);
      if (!dyptr->bptr) ErrorExit (vaxc$errno, FI_LI);
   }

   zptr = (sptr = dyptr->bptr) + dyptr->bsize;
   sptr += dyptr->blen;

   va_start (argptr, TextPtr);
   for (--argcnt; argcnt; argcnt--)
   {
      if (TextPtr)
      {
         cptr = TextPtr;
         TextPtr = NULL;
      }
      else
         cptr = (char*)va_arg (argptr, char*);
      if (!cptr) cptr = "{null}";
      while (*cptr)
      {
         if (sptr >= zptr)
         {
            dyptr->blen = sptr - dyptr->bptr;
            dyptr->bsize += dyptr->csize;
            if (dyptr->vmem)
               dyptr->bptr = CgiLibVeeMemRealloc (dyptr->bptr,
                                                  dyptr->bsize+32);
            else
               dyptr->bptr = realloc (dyptr->bptr,
                                      dyptr->bsize+32);
            if (!dyptr->bptr) ErrorExit (vaxc$errno, FI_LI);
            zptr = (sptr = dyptr->bptr) + dyptr->bsize;
            sptr += dyptr->blen;
         }
         *sptr++ = *cptr++;
      }
   }
   *sptr = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", dyptr->bptr);
   va_end (argptr);

   dyptr->blen = sptr - dyptr->bptr;
   return (dyptr->bptr);
}

/*****************************************************************************/
/*
Create a temporary mailbox (that can persist as long as the VWcms image for
CGIplus) to use for subprocess SYS$OUTPUT.  Spawn a NOWAIT subprocess and then
read its output until it exits.  The output is stored in a dynamic buffer and
returned as a pointer if a parameter is provided.
*/

int SpawnCommand
(
char *DclCommand,
char **OutputPtrPtr
)
{
#define SPAWN_MBX_BUFFER 2048
#define SPAWN_DATA_CHUNK 2048

   static unsigned long  SpawnFlags = 0x81;
   static unsigned short  MbxChan;
   $DESCRIPTOR (MbxNameDsc, "VWCMS_OUTPUT");

   int  status,
        CompletionStatus = 0,
        DataLength = 0,
        DataSize = 0;
   char  *cptr, *sptr, *zptr,
         *DataPtr = NULL;
   char  OutBuffer [SPAWN_MBX_BUFFER];
   struct {
      unsigned short  status;
      unsigned short  count;
      unsigned long  pid;
   } IOsb;
   $DESCRIPTOR (CommandDsc, DclCommand);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SpawnCommand()\n");

   if (!DclCommand)
   {
      sys$wake (0, 0);
      return (SS$_NORMAL);
   }

   if (WatchEnabled) WatchThis (FI_LI, "SPAWN !AZ", DclCommand);

   if (OutputPtrPtr) *OutputPtrPtr = NULL;

   if (!VWcmsConfig.SiteSpawn)
   {
      if (!strsame (VWcmsConfig.SiteName, "vdub", -1))
         return (SS$_NOPRIV);
      if (!TrnLnm ("VWCMS_SPAWN_VDUB", 0, 0))
         return (SS$_NOPRIV);
   }

   CommandDsc.dsc$w_length = strlen(DclCommand);

   if (!MbxChan)
   {
      /* if VWcms is CGIplus this mailbox will persist with the image */
      status = sys$crembx (0, &MbxChan, sizeof(OutBuffer),
                           sizeof(OutBuffer)-1, 0, 0, &MbxNameDsc, 0, 0);
      if (VMSnok(status)) return (status);
   }

   status = lib$spawn (&CommandDsc, 0, &MbxNameDsc, &SpawnFlags,
                       0, 0, &CompletionStatus, 0,
                       &SpawnCommand, 0, 0, 0, 0);
   if (WatchEnabled)
      WatchThis (FI_LI, "SPAWN !AZ (%X!8XL)", SysGetMsg(status), status);
   if (VMSnok (status)) return (status);

   zptr = (sptr = DataPtr) + DataSize;
   for (;;)
   {
      status = sys$qiow (0, MbxChan, IO$_READVBLK, &IOsb, 0, 0,
                         OutBuffer, sizeof(OutBuffer)-1, 0, 0, 0, 0);
      if (VMSok(status)) status = IOsb.status;
      if (VMSnok(status)) break;

      if (!OutputPtrPtr) continue;

      OutBuffer[IOsb.count] = '\0';
      for (cptr = OutBuffer; *cptr; cptr++)
      {
         if (sptr >= zptr)
         {
            DataLength = sptr - DataPtr;
            DataSize += SPAWN_DATA_CHUNK;
            DataPtr = CgiLibVeeMemRealloc (DataPtr, DataSize+32);
            if (!DataPtr) ErrorExit (vaxc$errno, FI_LI);
            zptr = (sptr = DataPtr) + DataSize;
            sptr += DataLength;
         }
         *sptr++ = *cptr;
      }
      *sptr++ = '\n';
   }
   *sptr = '\0';

   if (OutputPtrPtr) *OutputPtrPtr = DataPtr;

   if (status == SS$_ENDOFFILE) status = SS$_NORMAL;

   /* wait for the completion AST wake */
   sys$hiber ();

   if (WatchEnabled)
      WatchThis (FI_LI, "SPAWN !AZ (%X!8XL)",
                 SysGetMsg(CompletionStatus), CompletionStatus);

   return (CompletionStatus ? CompletionStatus : status);
}

/*****************************************************************************/
/*
Take out a lock on the specified resource.  Hash is elementary but should be
sufficient for purpose.  Hash function based on code by Arash Partow. 
*/

int WaitForLock
(
char *ResName,
BOOL ExMode
)
{
   static char  EnqName [32];
   static $DESCRIPTOR (EnqNameDsc, EnqName);

   unsigned int  x, hash, status;
   unsigned long  EnqMode;
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   hash = 0;
   for (cptr = ResName; *cptr; cptr++)
   {
      /* for this make it a case-insensitive hash */
      hash = (hash << 4) + (toupper(*cptr));
      if ((x = hash & 0xF0000000L) != 0)
      {
         hash ^= (x >> 24);
         hash &= ~x;
      }
   }
   hash &= 0x7FFFFFFF;

   EnqNameDsc.dsc$w_length = sprintf (EnqName, "VWCMS_%08.08X", hash);

   if (ExMode)
      EnqMode = LCK$K_EXMODE;
   else
      EnqMode = LCK$K_CRMODE;

   if (WatchEnabled) WatchThis (FI_LI, "ENQ !AZ !AZ", ResName, EnqName);

   sys$setprv (1, &SysLckMask, 0, 0);

   status = sys$enqw (0, EnqMode, &WaitForLockSb, LCK$M_SYSTEM,
                      &EnqNameDsc, 0, 0, 0, 0, 2, 0);

   sys$setprv (0, &SysLckMask, 0, 0);

   if (Debug) fprintf (stdout, "sys$enqw() %%X%08.08X\n", status);

   return (status);
}

/*****************************************************************************/
/*
Release the above lock.
*/

void WaitForUnLock ()

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WaitForUnLock()\n");

   status = sys$deq (WaitForLockSb.lksb$l_lkid, 0, 0, 0);
   if (Debug) fprintf (stdout, "sys$deq() %%X%08.08X\n", status);
}

/*****************************************************************************/
/*
Duplicate the supplied null-terminated string using CgiLibVeeMemCalloc().
*/

char* StringDup (char *sptr)

{
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "StringDup()\n");

   cptr = CgiLibVeeMemCalloc (strlen(sptr));
   if (!cptr) ErrorExit (vaxc$errno, FI_LI);
   strcpy (cptr, sptr);
   return (cptr);
}

/*****************************************************************************/
/*
Return a pointer to a string containing the lib$stat.. data.
*/

char* StatTimer (BOOL ShowStat)

{
   static $DESCRIPTOR (FaoDsc,
"REAL:!%T CPU:!UL.!2ZL DIO:!UL BIO:!UL FAULTS:!UL\0");

   static char  StatString [96];
   static $DESCRIPTOR (StatStringDsc, StatString);

   static unsigned long  LibStatTimerReal = 1,
                         LibStatTimerCpu = 2,
                         LibStatTimerBio = 3,
                         LibStatTimerDio = 4,
                         LibStatTimerFaults = 5;

   int  status;
   unsigned long  CpuBinTime,
                  CountBio,
                  CountDio,
                  CountFaults;
   unsigned long  RealBinTime [2];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "StatTimer()\n");

   if (!ShowStat)
   {
      lib$init_timer (0);
      return (NULL);
   }

   /* post-processing reset */
   lib$stat_timer (&LibStatTimerReal, &RealBinTime, 0);
   lib$stat_timer (&LibStatTimerCpu, &CpuBinTime, 0);
   lib$stat_timer (&LibStatTimerBio, &CountBio, 0);
   lib$stat_timer (&LibStatTimerDio, &CountDio, 0);
   lib$stat_timer (&LibStatTimerFaults, &CountFaults, 0);
   sys$fao (&FaoDsc, 0, &StatStringDsc,
            &RealBinTime, CpuBinTime/100, CpuBinTime%100,
            CountDio, CountBio, CountFaults);

   return (StatString);
}

/*****************************************************************************/
/*
Print to string/<stdout> using $FAO directives.
*/

int PrintFao
(
char *PrintBuffer,
int SizeOfPrintBuffer,
char *FaoString,
...
)
{
   int  argcnt, status;
   char  LocalBuffer [4096];
   unsigned short  ShortLength;
   unsigned long  *vecptr;
   unsigned long  FaoVector [64];
   $DESCRIPTOR (BufferDsc, LocalBuffer);
   $DESCRIPTOR (FaoStringDsc, FaoString);
   va_list  argptr;

   /*********/
   /* begin */
   /*********/

   va_count (argcnt);

   if (Debug)
      fprintf (stdout, "PrintFao() %d %d |%s|\n",
               SizeOfPrintBuffer, argcnt, FaoString);

   if (!FaoString) ErrorExit (SS$_BUGCHECK, FI_LI);
   if (PrintBuffer && SizeOfPrintBuffer <= 0) ErrorExit (SS$_BUGCHECK, FI_LI);
   if (argcnt > 64) ErrorExit (SS$_BUGCHECK, FI_LI);

   vecptr = FaoVector;
   va_start (argptr, FaoString);
   for (argcnt -= 3; argcnt; argcnt--)
      *vecptr++ = (unsigned long)va_arg (argptr, unsigned long);
   va_end (argptr);

   if (PrintBuffer)
   {
      BufferDsc.dsc$a_pointer = PrintBuffer;
      BufferDsc.dsc$w_length = SizeOfPrintBuffer-1;
   }
   FaoStringDsc.dsc$w_length = strlen(FaoString);
   status = sys$faol (&FaoStringDsc, &ShortLength, &BufferDsc,
                      (ULONGPTR)&FaoVector);
   if (Debug) fprintf (stdout, "sys$fao() %%X%08.08X\n", status);

   if (!(status & 1) || status == SS$_BUFFEROVF)
   {
      if (BufferDsc.dsc$w_length >= 20)
         ShortLength = sprintf (BufferDsc.dsc$a_pointer,
                                "$FAO() %%X%08.08X\n", status); 
      else
         PrintBuffer[ShortLength = 0] = '\0';
   }
   else
      PrintBuffer[ShortLength] = '\0';

   if (!PrintBuffer)
      fputs (LocalBuffer, stdout);
   else
   if (Debug)
      fprintf (stdout, "|%s|\n", PrintBuffer);

   return (ShortLength);
}

/*****************************************************************************/
/*
Exit after reporting the source code module name and line number.
*/

/* this is strictly a no-no */
extern int CgiLib__Environment;

ErrorExit
(
int StatusValue,
char *SourceName,
int SourceLine
)
{
   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (CgiLib__Environment)
      cptr = CgiLibVarNull ("SERVER_SOFTWARE");
   else
      cptr = NULL;

   if (cptr)
   {
      /* running as a CGI, report to client */
      CgiLibResponseHeader (403, "text/html",
                            "Last-Modified: %s\n", LastModifiedNow);

      fprintf (stdout,
"<html>\n\
<head>\n\
<meta name=\"generator\" content=\"%s\">\n\
<meta name=\"environment\" content=\"%s\">\n\
<meta name=\"copyright\" content=\"%s - GPL licensed\">\n\
<title>%s</title>\n\
</head>\n\
<body>\n\
<b>%s:</b>&nbsp; %s.\n\
<!-- ***** REPORTING MODULE:LINE IS %s:%d ***** --> \n\
</body>\n\
</html>\n",
               SoftwareEnv,
               CgiEnvironmentPtr,
               SoftwareCopy,
               "VWCMS",
               "Fatal VWcms Error",
               SysGetMsg(StatusValue),
               SourceName, SourceLine);
      exit (SS$_NORMAL);
   }

   fprintf (stdout, "%%%s-E-EXIT, %%X%08.08X %s:%d\n",
            Utility, StatusValue, SourceName, SourceLine);
   exit (StatusValue);
}

/*****************************************************************************/
/*
Hmmm, idea look familiar? :-)
*/

int WatchThis
(
char *SourceModule,
int SourceLine,
char *FaoString,
...
)
{
   static int  WatchCount;

   int  argcnt, status;
   char  *cptr, *sptr, *zptr;
   char  Buffer [16384],
         FaoBuffer [256],
         WhereAt [64];
   unsigned long  *vecptr;
   $DESCRIPTOR (BufferDsc, Buffer);
   $DESCRIPTOR (FaoBufferDsc, FaoBuffer);
   unsigned long  FaoVector [32];
   va_list  argptr;

   /*********/
   /* begin */
   /*********/

   va_count (argcnt);

   if (Debug)
     fprintf (stdout, "WatchThis() %s:%d %d |%s|\n",
              SourceModule, SourceLine, argcnt, FaoString);

   if (!FaoString)
   {
      if (WatchCount)
      {
         /* post-processing reset */
         WatchThis (FI_LI, "STATS !AZ", StatTimer(TRUE));
         WatchThis (FI_LI, "WATCH end");
         WatchEnabled = WatchCount = 0;
      }
      else
      {
         /* pre-processing initialize */
         WatchCount = 1;
         CgiLibResponseHeader (200, "text/plain",
                               "Last-Modified: %s\n"
                               "Script-Control: X-content-encoding-gzip=0\n",
                               LastModifiedNow);
         WatchThis (FI_LI, "WATCH begin");
      }
      return (SS$_NORMAL);
   }

   sprintf (WhereAt, "%s:%d", SourceModule, SourceLine);

   vecptr = FaoVector;
   *vecptr++ = 0;
   *vecptr++ = (unsigned long)WhereAt;
   va_start (argptr, FaoString);
   for (argcnt -= 1; argcnt; argcnt--)
      *vecptr++ = (unsigned long)va_arg (argptr, unsigned long);
   va_end (argptr);

   zptr = (sptr = FaoBuffer) + sizeof(FaoBuffer)-3;
   for (cptr = "|!%T|!12AZ|"; *cptr; *sptr++ = *cptr++);
   for (cptr = FaoString; *cptr && sptr < zptr; *sptr++ = *cptr++);
   *sptr++ = '|';
   *sptr++ = '\n';
   *sptr++ = '\0';
   FaoBufferDsc.dsc$a_pointer = FaoBuffer;
   FaoBufferDsc.dsc$w_length = sptr - FaoBuffer;
   status = sys$faol (&FaoBufferDsc, 0, &BufferDsc,
                      (ULONGPTR)&FaoVector);
   if (Debug) fprintf (stdout, "sys$fao() %%X%08.08X\n", status);
   if (!(status & 1)) exit (status);
   fputs (Buffer, stdout);

   return (status);
}

/*****************************************************************************/
/*
Ditto? :-)
*/

int WatchDump
(
char *DataPtr,
int DataLength
)
{
   int  cnt, len;
   char  *cptr, *lptr, *sptr;
   char  WatchBuffer [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "WatchDump() %d\n", DataLength);

   cptr = DataPtr;
   len = DataLength;
   while (len)
   {
      sptr = WatchBuffer;
      lptr = cptr;
      cnt = 0;
      while (cnt < 32)
      {
         if (cnt < len)
            sptr += sprintf (sptr, "%02.02X", (unsigned char)*cptr++);
         else
         {
            *sptr++ = ' ';
            *sptr++ = ' ';
         }
         if (!(++cnt % 4)) *sptr++ = ' ';
      }
      *sptr++ = ' ';
      cptr = lptr;
      cnt = 0;
      while (cnt < 32 && cnt < len)
      {
         if (isprint(*cptr))
            *sptr++ = *cptr;
         else
            *sptr++ = '.';
         cptr++;
         cnt++;
      }
      if (len > 32) len -= 32; else len = 0;
      *sptr++ = '\n';
      *sptr = '\0';
      fputs (WatchBuffer, stdout);
   }

   return (1);
}

/*****************************************************************************/
/*
When 'FileSpec' is non-NULL initiate a new file-system search.  With each
subsequent call return the specification of the next hit in the file-system
search.
*/

char* FileSpecSearch
(
int NameOnly,
char *DirSpec,
char *FileSpec
)
{
   static BOOL  VersionWildcard;
   static char  ExpFileName [255+1],
                RetFileName [255+1],
                ResFileName [255+1];
   static struct FAB  SearchFab;
   static struct NAM  SearchNam;

   int  status;
   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "FileSpecSearch()\n");

   if (FileSpec)
   {
      SearchFab = cc$rms_fab;
      SearchFab.fab$l_dna = DirSpec;
      SearchFab.fab$b_dns = strlen(DirSpec);
      SearchFab.fab$l_fna = FileSpec;
      SearchFab.fab$b_fns = strlen(FileSpec);
      SearchFab.fab$l_nam = &SearchNam;

      SearchNam = cc$rms_nam;
      SearchNam.nam$l_esa = ExpFileName;
      SearchNam.nam$b_ess = sizeof(ExpFileName)-1;
      SearchNam.nam$l_rsa = ResFileName;
      SearchNam.nam$b_rss = sizeof(ResFileName)-1;

      if (VMSnok (status = sys$parse (&SearchFab, 0, 0)))
      {
         if (Debug) fprintf (stdout, "sys$parse() %%X%08.08X\n", status);
         sprintf (ExpFileName, "%X08.08X", status);
         return (ExpFileName);
      }
      /* if no version component in the specification */
      if (strstr (FileSpec, ";*"))
         VersionWildcard = TRUE;
      else
         VersionWildcard = FALSE;
      SearchNam.nam$l_ver[SearchNam.nam$b_ver] = '\0';
      if (Debug) fprintf (stdout, "|%s|\n", ExpFileName);
      return (ExpFileName);
   }

   if (VMSok (status = sys$search (&SearchFab, 0, 0)))
   {
      SearchNam.nam$l_ver[SearchNam.nam$b_ver] = '\0';
      if (Debug) fprintf (stdout, "|%s|\n", ResFileName);
      sptr = RetFileName;
      if (NameOnly)
      {
         for (cptr = ResFileName; *cptr; cptr++);
         while (cptr > ResFileName && *cptr != ']') cptr--;
         if (*cptr == ']')
         {
            cptr++;
            while (*cptr) *sptr++ = *cptr++;
            *sptr = '\0';
         }
         else
            ErrorExit (SS$_BUGCHECK, FI_LI);
      }
      else
      {
         for (cptr = ResFileName; *cptr; *sptr++ = *cptr++);
         *sptr = '\0';
      }
      if (!VersionWildcard)
      {
         while (sptr > RetFileName && *sptr != ';') sptr--;
         *sptr = '\0';
      }
      return (RetFileName);
   }
   if (Debug) fprintf (stdout, "sys$search() %%X%08.08X\n", status);
   return (NULL);
}

/*****************************************************************************/
/*
Translate a logical name using LNM$FILE_DEV or LNM$SYSTEM.  Returns a pointer
to the value string, or NULL if the name does not exist.  If 'LogValue' is
supplied the logical name is translated into that (assumed to be large enough),
otherwise it's translated into an internal static buffer.  'IndexValue' should
be zero for a 'flat' logical name, or 0..127 for interative translations.
*/

char* TrnLnm
(
char *LogName,
char *LogValue,
int IndexValue
)
{
   static unsigned short  ValueLength;
   static unsigned long  LnmAttributes,
                         LnmIndex;
   static char  StaticLogValue [256];
   static $DESCRIPTOR (LogNameDsc, "");
   static $DESCRIPTOR (LnmTableDsc, "LNM$FILE_DEV");
/***
   static $DESCRIPTOR (LnmTableDsc, "LNM$SYSTEM");
***/
   static struct {
      short int  buf_len;
      short int  item;
      void  *buf_addr;
      unsigned short  *ret_len;
   } LnmItems [] =
   {
      { sizeof(LnmIndex), LNM$_INDEX, &LnmIndex, 0 },
      { sizeof(LnmAttributes), LNM$_ATTRIBUTES, &LnmAttributes, 0 },
      { 255, LNM$_STRING, 0, &ValueLength },
      { 0,0,0,0 }
   };

   int  status;
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TrnLnm() |%s| %d\n", LogName, IndexValue);

   LnmIndex = IndexValue;

   LogNameDsc.dsc$a_pointer = LogName;
   LogNameDsc.dsc$w_length = strlen(LogName);
   if (LogValue)
      cptr = LnmItems[2].buf_addr = LogValue;
   else
      cptr = LnmItems[2].buf_addr = StaticLogValue;

   status = sys$trnlnm (0, &LnmTableDsc, &LogNameDsc, 0, &LnmItems);
   if (Debug) fprintf (stdout, "sys$trnlnm() %%X%08.08X\n", status);
   if (!(status & 1) || !(LnmAttributes & LNM$M_EXISTS))
   {
      if (LogValue) LogValue[0] = '\0';
      if (Debug) fprintf (stdout, "|(null)|\n");
      return (NULL);
   }

   cptr[ValueLength] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", cptr);
   return (cptr);
}

/****************************************************************************/
/*
Set an integer reflecting the major, minor version and update release of VMS
(e.g. 600, 610, 620...730, 731, 732, etc.)
*/ 

int GetVmsVersion ()

{
#define SYI$_VERSION 4096

   static char  SyiVersion [16];
   static VMS_ITEM_LIST3 SyiItems [] = {
      { 8, SYI$_VERSION, &SyiVersion, 0 },
      { 0,0,0,0 }
   };

   int  status, version;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "GetVmsVersion()\n");

   if (VMSnok (status = sys$getsyiw (0, 0, 0, &SyiItems, 0, 0, 0)))
      exit (status);
   SyiVersion[8] = '\0';
   if (SyiVersion[0] == 'V' && isdigit(SyiVersion[1]) &&
       SyiVersion[2] == '.' && isdigit(SyiVersion[3]))
   {
      /* e.g. "V7.3" */
      version = ((SyiVersion[1]-48) * 100) + (SyiVersion[3]-48) * 10;
      /* if something like "V7.3-2" */
      if (SyiVersion[4] == '-') version += SyiVersion[5]-48;
   }
   else
      /* unknown format, assume it's version 6.0 */
      version = 600;
   if (Debug) fprintf (stdout, "|%s| %d\n", SyiVersion, version);
   return (version);
}

/****************************************************************************/
/*
Return a pointer to a buffer containg a string describing the hardware and VMS
version.
*/ 

char* GetPlatformDescription ()

{
#define SYI$_HW_NAME 4362
#define SYI$_VERSION 4096

   static unsigned short  SyiHwNameLength;
   static char  Description [61+9+5],
                SyiHwName [61],
                SyiVersion [9];
   static VMS_ITEM_LIST3 SyiItems [] = {
      { sizeof(SyiHwName)-1, SYI$_HW_NAME, &SyiHwName, &SyiHwNameLength },
      { sizeof(SyiVersion)-1, SYI$_VERSION, &SyiVersion, 0 },
      { 0,0,0,0 }
   };

   int  status, version;
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "GetPlatformDescription()\n");

   if (Description[0]) return (Description);
   if (VMSnok (status = sys$getsyiw (0, 0, 0, &SyiItems, 0, 0, 0)))
      exit (status);
   SyiHwName[SyiHwNameLength] = '\0';
   SyiVersion[8] = '\0';
   for (cptr = SyiVersion; *cptr && *cptr != ' '; cptr++);
   *cptr = '\0';
   sprintf (Description, "%s VMS %s", SyiHwName, SyiVersion);
   return (Description);
}

/*****************************************************************************/
/*
*/

char* SysGetMsg (int StatusValue)

{
   static char  StatusMsg [256];

   int  status;
   unsigned short  Length;
   char  *cptr;
   $DESCRIPTOR (StatusMsgDsc, StatusMsg);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SysGetMsg() %%X%08.08X\n", StatusValue);

#  pragma message __save
#  pragma message disable (ADDRCONSTEXT)

   /* cater for VAX environment as best we can */
#  pragma message disable (IMPLICITFUNC)
#  ifdef __VAX
#     define SYS$GETMSG sys$getmsg
#  endif
   status = sys$getmsg (StatusValue, &Length, &StatusMsgDsc, 1, 0);
#  pragma message __restore
   if (!(status & 1)) ErrorExit (status, FI_LI);
   StatusMsg[Length] = '\0';
   StatusMsg[0] = toupper(StatusMsg[0]);
   if (cptr = strstr (StatusMsg, "!/!_")) *(ULONGPTR)cptr = '.   ';

   return (StatusMsg);
}

/******************************************************************************/
/*
Convert any data to base-64.
Allocates storage to parameter length plus 40% and populates.
Returns pointer to same, or NULL if there was some issue during the encoding.
*/

char* Base64Encode
(
char *DataPtr,
int DataLength
)
{
#define CHAR64(c) (Base64Chars[(unsigned char)(c)])

   int  ccnt,
        EncodedSize;
   unsigned char  c1, c2, c3;
   char  *cptr, *ccptr, *czptr, *sptr, *zptr,
         *EncodedPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "Base64Encode() %d\n", DataLength);

   /* base-64 requires about 40% more storage than the original data */
   if (DataLength < 0) DataLength = strlen(DataPtr);
   if (DataLength >= 128)
      EncodedSize = DataLength + (DataLength / 2);
   else
      EncodedSize = 256;
   if (Debug) fprintf (stdout, "%d\n", EncodedSize);
   EncodedPtr = CgiLibVeeMemCalloc (EncodedSize+16);
   if (!EncodedPtr) ErrorExit (vaxc$errno, FI_LI);

   zptr = (sptr = EncodedPtr) + EncodedSize;
   czptr = (cptr = DataPtr) + DataLength;

   while (cptr < czptr && sptr < zptr)
   {
      ccnt = 1;
      c1 = (unsigned char)*cptr++;
      if (cptr < czptr)
      {
         c2 = (unsigned char)*cptr++;
         ccnt++;
      }
      else
         c2 = 0;
      if (cptr < czptr)
      {
         c3 = (unsigned char)*cptr++;
         ccnt++;
      }
      else
         c3 = 0;

      *sptr++ = CHAR64 (c1>>2);
      *sptr++ = CHAR64 (((c1 & 0x3)<< 4) | ((c2 & 0xf0) >> 4));
      if (ccnt == 1)
      {
         *sptr++ = '=';
         *sptr++ = '=';
      }
      else
      if (ccnt == 2)
      {
         *sptr++ = CHAR64 (((c2 & 0x0f) << 2) | ((c3 & 0xc0) >> 6));
         *sptr++ = '=';
      }
      else
      {
        *sptr++ = CHAR64 (((c2 & 0xf) << 2) | ((c3 & 0xc0) >> 6));
        *sptr++ = CHAR64 (c3 & 0x3f);
      }
   }
   if (sptr >= zptr)
   {
      if (Debug) fprintf (stdout, "OVERFLOW!\n");
      return (NULL);
   }
   *sptr = '\0';

   return (EncodedPtr);

#undef CHAR64
}

/******************************************************************************/
/*
Convert any data to URL-safe base-64 (use '-' for of '+' and '_' for '/').
Allocates storage to parameter length plus 40% and populates.
Returns pointer to same, or NULL if there was some issue during the encoding.
*/

char* Base64UrlSafeEncode
(
char *DataPtr,
int DataLength
)
{
#define CHAR64(c) (Base64UrlSafeChars[(unsigned char)(c)])

   int  ccnt,
        EncodedSize;
   unsigned char  c1, c2, c3;
   char  *cptr, *ccptr, *czptr, *sptr, *zptr,
         *EncodedPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "Base64Encode() %d\n", DataLength);

   /* base-64 requires about 40% more storage than the original data */
   if (DataLength < 0) DataLength = strlen(DataPtr);
   if (DataLength >= 128)
      EncodedSize = DataLength + (DataLength / 2);
   else
      EncodedSize = 256;
   if (Debug) fprintf (stdout, "%d\n", EncodedSize);
   EncodedPtr = CgiLibVeeMemCalloc (EncodedSize+16);
   if (!EncodedPtr) ErrorExit (vaxc$errno, FI_LI);

   zptr = (sptr = EncodedPtr) + EncodedSize;
   czptr = (cptr = DataPtr) + DataLength;

   while (cptr < czptr && sptr < zptr)
   {
      ccnt = 1;
      c1 = (unsigned char)*cptr++;
      if (cptr < czptr)
      {
         c2 = (unsigned char)*cptr++;
         ccnt++;
      }
      else
         c2 = 0;
      if (cptr < czptr)
      {
         c3 = (unsigned char)*cptr++;
         ccnt++;
      }
      else
         c3 = 0;

      *sptr++ = CHAR64 (c1>>2);
      *sptr++ = CHAR64 (((c1 & 0x3)<< 4) | ((c2 & 0xf0) >> 4));
      if (ccnt == 1)
      {
         *sptr++ = '=';
         *sptr++ = '=';
      }
      else
      if (ccnt == 2)
      {
         *sptr++ = CHAR64 (((c2 & 0x0f) << 2) | ((c3 & 0xc0) >> 6));
         *sptr++ = '=';
      }
      else
      {
        *sptr++ = CHAR64 (((c2 & 0xf) << 2) | ((c3 & 0xc0) >> 6));
        *sptr++ = CHAR64 (c3 & 0x3f);
      }
   }
   if (sptr >= zptr)
   {
      if (Debug) fprintf (stdout, "OVERFLOW!\n");
      return (NULL);
   }
   *sptr = '\0';

   return (EncodedPtr);

#undef CHAR64
}

/*****************************************************************************/
/*
Generic base64 in-situ decode function.  Returns the size of the decoded data
or -1 and an empty string to indicate an error.
*/

int Base64Decode (char *Base64Ptr)

{
#define CHAR64(c) (Base64Table[c&0x7f])

   int  ccnt,
        BufferLength,
        ErrorLine;
   unsigned long  reg;
   unsigned char  val;
   char  *cptr, *sptr,
         *ErrorPtr = "";

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "Base64Decode()\n");

   if (!Base64Ptr) ErrorExit (SS$_BUGCHECK, FI_LI);
   cptr = sptr = Base64Ptr;

   while (*cptr)
   {
      reg = 0;
      if (*cptr)
      {
         if (*cptr == '=')
         {
            ErrorPtr = cptr;
            ErrorLine = __LINE__;
            goto MimeDecBase64_fail;
         }
         if ((val = CHAR64(*cptr++)) == XX)
         {
            ErrorPtr = cptr;
            ErrorLine = __LINE__;
            goto MimeDecBase64_fail;
         }
         reg |= val << 18;
      }
      if (*cptr)
      {
         if (*cptr == '=') goto MimeDecBase64_fail;
         ccnt = 1;
         if ((val = CHAR64(*cptr++)) == XX)
         {
            ErrorPtr = cptr;
            ErrorLine = __LINE__;
            goto MimeDecBase64_fail;
         }
         reg |= val << 12;
      }
      if (*cptr)
      {
         if (*cptr == '=')
            cptr++;
         else
         {
            ccnt++;
            if ((val = CHAR64(*cptr++)) == XX)
            {
               ErrorPtr = cptr;
               ErrorLine = __LINE__;
               goto MimeDecBase64_fail;
            }
            reg |= val << 6;
         }
      }
      if (*cptr)
      {
         if (*cptr == '=')
            cptr++;
         else
         {
            ccnt++;
            if ((val = CHAR64(*cptr++)) == XX)
            {
               ErrorPtr = cptr;
               ErrorLine = __LINE__;
               goto MimeDecBase64_fail;
            }
            reg |= val;
         }
      }
      else
         goto MimeDecBase64_fail;

      if (ccnt--) *sptr++ = (reg >> 16) & 0xff;
      if (ccnt--) *sptr++ = (reg >> 8) & 0xff;
      if (ccnt--) *sptr++ = reg & 0xff;

      while (*cptr && (*cptr == '\r' || *cptr == '\n')) cptr++;
   }

   *sptr = '\0';

   if (WatchEnabled)
      WatchThis (FI_LI, "BASE64 ok !UL bytes", sptr-Base64Ptr);

   return (sptr - Base64Ptr);

MimeDecBase64_fail:

   if (WatchEnabled)
      WatchThis (FI_LI, "BASE64 error; !UL |!#AZ",
                 ErrorLine,
                 strlen(ErrorPtr) > 160 ? 160 : strlen(ErrorPtr), ErrorPtr);

   *Base64Ptr = '\0';

   return (-1);

#undef CHAR64
}  

/****************************************************************************/
/*
Escape plain-text characters forbidden to occur in HTML documents (i.e. '<',
'>' and '&').
*/ 
 
char* HtmlEscape (char *StringPtr)

{
   int  ccnt;
   char  *cptr, *sptr,
         *EscapedPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HtmlEscape() |%s|\n", StringPtr);

   if (!StringPtr || !*StringPtr) return ("");

   ccnt = 0;
   for (cptr = StringPtr; *cptr; cptr++)
   {
      switch (*cptr)
      {
         case '<' : ccnt += 4; continue;
         case '>' : ccnt += 4; continue;
         case '&' : ccnt += 5; continue;
         case '"' : ccnt += 6; continue;
         default : ccnt++;
       }
   }
   ccnt++;
   EscapedPtr = sptr = CgiLibVeeMemCalloc (ccnt);
   if (!EscapedPtr) ErrorExit (vaxc$errno, FI_LI);

   for (cptr = StringPtr; *cptr; cptr++)
   {
      switch (*cptr)
      {
         case '<' :
            memcpy (sptr, "&lt;", 4); sptr += 4; continue;
         case '>' :
            memcpy (sptr, "&gt;", 4); sptr += 4; continue;
         case '&' :
            memcpy (sptr, "&amp;", 5); sptr += 5; continue;
         case '"' :
            memcpy (sptr, "&quot;", 6); sptr += 6; continue;
         default :
            *sptr++ = *cptr;
      }
   }
   *sptr = '\0';
   if (Debug) fprintf (stdout, "%d |%s|\n", sptr-EscapedPtr, EscapedPtr);

   return (EscapedPtr);
}
 
/****************************************************************************/
/*
Escape characters for inside the XML of an RSS feed.
*/ 
 
char* RssEscape (char *StringPtr)

{
   int  ccnt;
   char  *cptr, *sptr,
         *EscapedPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "RssEscape() |%s|\n", StringPtr);

   if (!StringPtr || !*StringPtr) return ("");

   ccnt = 0;
   for (cptr = StringPtr; *cptr; cptr++)
   {
      switch (*cptr)
      {
         case '<' : ccnt += 8; continue;
         case '>' : ccnt += 8; continue;
         case '&' : ccnt += 5; continue;
         default : ccnt++;
       }
   }
   ccnt++;
   EscapedPtr = sptr = CgiLibVeeMemCalloc (ccnt);
   if (!EscapedPtr) ErrorExit (vaxc$errno, FI_LI);

   for (cptr = StringPtr; *cptr; cptr++)
   {
      switch (*cptr)
      {
         case '<' :
            memcpy (sptr, "&amp;lt;", 8); sptr += 8; continue;
         case '>' :
            memcpy (sptr, "&amp;gt;", 8); sptr += 8; continue;
         case '&' :
            memcpy (sptr, "&amp;", 5); sptr += 5; continue;
         default :
            *sptr++ = *cptr;
      }
   }
   *sptr = '\0';
   if (Debug) fprintf (stdout, "%d |%s|\n", sptr-EscapedPtr, EscapedPtr);

   return (EscapedPtr);
}
 
/*****************************************************************************/
/*
*/

char* TidyTextArea
(
char *TextPtr,
int WrapAt
)
{
   int  cnt;
   char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TidyTextArea()\n");

   /* munge those pesky <CR>s from <CR><LF>s */
   for (cptr = sptr = TextPtr; *cptr; *sptr++ = *cptr++)
      if (*(USHORTPTR)cptr == '\r\n') cptr++;
   *sptr = '\0';

   /* tidy leading newlines */
   while (*TextPtr == '\n') TextPtr++;

   /* if the text is empty return a null string */
   for (cptr = TextPtr; isspace(*cptr); cptr++);
   if (!*cptr) return (cptr);

   /* tidy trailing white-space */
   while (*cptr) cptr++;
   if (cptr > TextPtr) cptr--;
   while (cptr > TextPtr && isspace(*cptr)) cptr--;
   if (cptr > TextPtr) cptr++;
   *cptr = '\0';

   /* add any line wrap */
   cptr = TextPtr;
   while (*cptr)
   {
      while (*cptr == '\n') cptr++;
      if (!*cptr) break;

      sptr = cptr;
      while (*cptr == ' ') cptr++;
      if (cptr - sptr >= 2)
      {
         /* do not wrap lines beginning with two or more spaces */
         while (*cptr && *cptr != '\n') cptr++;
         continue;
      }

      cnt = 0;
      sptr = NULL;
      while (*cptr && *cptr != '\n')
      {
         if (*cptr == ' ' || *cptr == '\t')
           for (sptr = cptr; *(sptr+1) == ' ' || *(sptr+1) == '\t'; sptr++);
         if (++cnt > WrapAt && sptr)
         {
            /* add line wrap to the last white-space previously noted */
            *sptr = '\n';
            sptr = NULL;
            cnt = 0;
         }
         if (*cptr == '\n')
         {
            sptr = NULL;
            cnt = 0;
         }
         cptr++;
      }
   }

   return (TextPtr);
}

/*****************************************************************************/
/*
It just confines the instances of the warning issue described immediately above
the #include netdb.h above.
*/

char* HostNameByAddr (char *IpAddrStr)

{
   int  IpAddr;
   char  *cptr;
   struct hostent  *HostEntryPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HostNameByAddr()\n");

   IpAddr = inet_addr (IpAddrStr);
   if (IpAddr == -1) return (NULL);
   HostEntryPtr = gethostbyaddr (&IpAddr, 4, AF_INET);
   if (!HostEntryPtr) return (NULL);
   cptr = CgiLibVeeMemCalloc (strlen(HostEntryPtr->h_name)+1);
   if (!cptr) ErrorExit (vaxc$errno, FI_LI);
   strcpy (cptr, HostEntryPtr->h_name);
   return (cptr); 
}

/****************************************************************************/
/*
Copies null-terminated string 'SrcPtr' to 'DestPtr'.
Returns length of copied string.
Truncate on overflow!
*/ 
int strzcpy
(
char *DestPtr,
char *SrcPtr,
int SizeOfDest
)
{
   char  *cptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   zptr = (sptr = DestPtr) + SizeOfDest-1;
   for (cptr = SrcPtr; *cptr && sptr < zptr; *sptr++ = *cptr++);
   *sptr = '\0';
   return (sptr - DestPtr);
}

/****************************************************************************/
/*
Does a case-insensitive, character-by-character string compare and returns 
true if two strings are the same, or false if not.  If a maximum number of 
characters are specified only those will be compared, if the entire strings 
should be compared then specify the number of characters as 0.
*/ 

BOOL strsame
(
char *sptr1,
char *sptr2,
int  count
)
{
   /*********/
   /* begin */
   /*********/

   while (*sptr1 && *sptr2)
   {
      if (toupper (*sptr1++) != toupper (*sptr2++)) return (FALSE);
      if (count)
         if (!--count) return (TRUE);
   }
   if (*sptr1 || *sptr2)
      return (FALSE);
   else
      return (TRUE);
}

/*****************************************************************************/

