/*****************************************************************************/
/*
                               strip_BOM_CR.c

This utility strips the UTF-* BOM (byte order mark)

  http://unicode.org/faq/utf_bom.html#BOM

and the <CR> of the DOS carriage-control <CR><LF> sequence.

I'm not sure why FCKeditor is released using this format but it's disruptive to
our purposes unless removed.  Perhaps out VMSTAR and/or UNZIP should remove it?

The second is not absolutely necessary but if we're going to the trouble of
removing the BOM we might as well tidy the carriage-control as well!

Of course, this process is only necessary when bringing over a new version of
FCKeditor.

To build:

  $ CC STRIP_BOM_CR
  $ LINK STRIP_BOM_CR

From the FCKeditor directory invoke it using

  $ MCR [wherever]STRIP_BOM_CR [...]*.js
  $ MCR [wherever]STRIP_BOM_CR [...]*.html
  $ MCR [wherever]STRIP_BOM_CR [...]*.xml
  $ MCR [wherever]STRIP_BOM_CR [...]*.txt
  $ PURGE [...]


COPYRIGHT
---------
Copyright (C) 2008-2011 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it under the
conditions of the GNU GENERAL PUBLIC LICENSE, version 3, or any later version.
http://www.gnu.org/licenses/gpl.txt


VERSION HISTORY
---------------
24-MAR-2006  MGD  initial
*/
/*****************************************************************************/

#include <ctype.h>
#include <errno.h>
#include <stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unixio.h>
#include <unixlib.h>

int stripCount;

int stripBOMcr (char*);

/*****************************************************************************/
/*
*/

int main (int argc, char* argv[])

{
   int  retval;
   char  inName [512];

   strcpy (inName, argv[1]);
   if (!strchr (inName, ';')) strcat (inName, ";0");

   retval = decc$from_vms (inName, stripBOMcr, 1);

   printf ("%d files processed, %d stripped\n", retval, stripCount);
}

/*****************************************************************************/
/*
*/

int stripBOMcr (char* fName)

{
   static char  bom [3] = { 0xef, 0xbb, 0xbf };

   int  ccnt, retval;
   char  *aptr, *cptr, *sptr;
   char  outName [512];
   FILE  *inf, *outf;
   stat_t  statBuf;

   printf ("%s\n", fName);

   if (stat (fName, &statBuf) != 0) exit (vaxc$errno);
                                    
   aptr = cptr = calloc (1, (ccnt=statBuf.st_size)+32);
   if (!aptr) exit (vaxc$errno);

   if (!(inf = fopen (fName, "r"))) exit (vaxc$errno);

   if ((retval = read (fileno(inf), aptr, ccnt)) < 0) exit (vaxc$errno);

   fclose (inf);

   sptr = cptr;
   while (*cptr)
   {
      if (*(USHORTPTR)cptr == '\r\n') cptr++;
      *sptr++ = *cptr++;
   }

   strcpy (outName, fName);
   for (cptr = outName; *cptr; cptr++);
   if (cptr > outName) cptr--;
   while (isdigit(*cptr)) cptr--;
   if (*cptr == '.') *cptr = '\0';
   strcat (outName, ";0");

   if (!(outf = fopen (outName, "w"))) exit (vaxc$errno);

   cptr = aptr;
   if (cptr == sptr && memcmp (cptr, bom, sizeof(bom)))
   {
      free (aptr);
      return (1);
   }
   if (!memcmp (cptr, bom, sizeof(bom))) cptr += 3;

   if (write (fileno(outf), cptr, sptr-cptr) <= 0) exit (vaxc$errno);

   fclose (outf);

   free (aptr);

   stripCount++;

   return (1);
}

/*****************************************************************************/
