/*****************************************************************************/
/*
                                   VWcms.c

VMS Web-content management system!!
~   ~   ~       ~          ~

Affectionately known as the Vee-Dub[-See-Em-Ess] and written "Vdub".

A request for content can contain the following combination of components:

  /site/
  /site/section
  /site/section/subsection
  /site/section/../subsection

The depth of subsections can be determined by configuration directive
[site-depth] (3 by default).

Although not mandatory (provided the web server has permission VWcms can access
a site wherever it is located in the file-system) it is recommended that
multiple site all be located under a single root using (for example)

  $ DEFINE /SYSTEM /TRANSLATION=CONCEALED VWCMS_ROOT DKA0:[VWCMS.]

The site's file are located using the GLOBAL configuration directive
[site-directory].  This directive MUST be configured.

The site content file is by default _INDEX.HTML.  The GLOBAL configuration
directive [site-content] can change this (the underscore is not required).


DATA CACHE
----------
For WASD CGIplus file content is cached.  The number of concurrent entries can
be explicitly specified using the logical name 

   $ DEFINE /SYSTEM VWCMS_DATA_CACHE_MAX 128

The maximum size of any one file is 256kB by default.  This can be explicitly
specified using the logical name

   $ DEFINE /SYSTEM VWCMS_DATA_CACHE_SIZE_MAX 1024


CONTENT CACHE
-------------
For WASD CGIplus parsed content is cached.  The maximum number of concurrent
entries by default is set to one eighth of those in the data cache.  This
number can be explicitly specified using the logical name

   $ DEFINE /SYSTEM VWCMS_CONTENT_CACHE_MAX 16


CONFIGURATION
-------------
See description in CONFIG.C

For system-wide application the logical name VWCMS_DOWNTIME parallels the
[downtime] directive.


MEMORY MANAGEMENT
-----------------
This application uses the CgiLibVeeMem..() functions to manage application
allocated memory and assist in preventing leaks.  Some memory is still
calloc()ed (especially by CGILIB) but that is the more straight-forward to keep
track of.  Essentially this allows an application to 'forget' about freeing
allocated memory and have a VM zone reset take care of it all after-the-event.


COMMAND-LINE
------------
/DBUG          enable debug statements
/VERSION       display version and copyright details
/WATCH         enable WATCH statements


COPYRIGHT
---------
Copyright (C) 2008-2011 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it under the
conditions of the GNU GENERAL PUBLIC LICENSE, version 3, or any later version.
http://www.gnu.org/licenses/gpl.txt


VERSION HISTORY
---------------
29-AUG-2011  MGD  v1.2.1,
                  ..Present() if not found in content look in template
                  bugfix; OutputFromTo() leading '/' processing (on root site)
31-JUL-2011  MGD  v1.2.0,
                  Google reCAPTCHA (with acknowledgement to Tim Sneddon for the
                    excellent idea but who was beaten to the punch by insomnia)
                  Question-and-answer (QandA) submission challenge
                  CommentSubmitFind() look in section content then if not
                    found in template content (in case of site-wide comment)
                  bugfix; ManageContentHtmlEditor() insert image path
                  bugfix; ContentLoad() .htc URL
                  bugfix; TicketFile()
18-MAY-2011  MGD  v1.1.0, hot on the heels
                  site management copy/remove/insert section trees
                  $$TAG$cloud=..$ and $$TAG$list=.." goto=<section> parameter
                  ContentParseLine1(), TemplateLoad() and TemplateContent()
                    nested conditional testing (per TES) 
                  CommentRejectString() and CommentRejectWord()
                    reject all comments if "*"
                  $$IF$query=<keyword>$$ tests query string keyword
                  $$INCLUDE$$ and $$INSERT$$ refresh=<integer>
                  [comment-reject-agent] configuration directive
                  bugfix; ManageResources() "<img src=/_..>"
                  bugfix; ContentTocCss() hierarchy 'more' detection
                  bugfix; SynopsisContent() XML item RssEscape() (per TES)
                  bugfix; nested conditional testing (per TES)
                  bugfix; ContentProcess() RSS query processing (per TES)
                  bugfix; ConfigLoad() repoint after ConfigConcat()
09-APR-2011  MGD  v1.0.0, ta-da!
                  draggable and closable management interface
                  $$MANAGE$absoloute=X,Y$$ to position management interface
                  $$BLOG$new$$ to generate dialog for new/next blog entry
                  [PMDF-kludge] to work-around PMDF sender quirk (e.g. IN%"")
                  ManageCheckSectionName() to eliminate duplicate sections
                  ManageChangeSectionName() to globally change section links
                  bugfix; ConfigMatch() wildcard match
                  bugfix; FormSubmitSpawn() fopen/close/delete than "fop=dlt"
02-MAY-2010  MGD  [site-email-from] global config specifies email origin
                  $$BLAH$$origin=..$$ parameter specifies email origin
                  OutputFromTo() two successive '_' means don't touch!
                  ConfigTextParse() allow "[<spaces>name] value" 
15-APR-2010  MGD  CAUTION: site configuration has moved from using logical
                           names to using the VWCMS_CONFIG file
                  ConfigGlobalLoad() to load system-wide VWcms configuration
                  TemplateInclude() significant rework
                  regular expression searching and conditional matching
                  $$INCLUDE and $$INSERT$$ have revised syntax
                  $$NOPRINT$$ and $$NOPRINT$+$ suppresses section printing
                  $$TAG$$ allows tagging sections
                  $$BLOG$latest=..$$ 
26-JAN-2010  MGD  bugfix; TemplateTestIf()
17-JAN-2010  MGD  $$FORM$$ processing
                  $$TICKET$$ and associated TicketFile()
                  ManageSecure() applied at [publish] and [tidy]
                  [ods] configuration and MakeFileName()
28-NOV-2009  MGD  $$HTML$$ allows 'raw' HTML content
                  add "validate=" and "validator=" to $$RSS$$
                  [reject-agent] (primarily for SPAMbot handling)
                  $$IF$$ string operators (per TES)
                  bugfix; TemplateTestIf() (per TES)
30-OCT-2009  MGD  the encoding of a forward-slash within a section name
                    into blackslash is explicitly processed so it appears
                    as a forward-slash everywhere except in the HTML source
                  prototype decc$write_eof_to_mbx() for VAX (per TES)
26-SEP-2009  MGD  $$RSS$$
                  add [lang-rss]
                  $$NOTOC$$ becomes $$HIDE$$ and $$NOTOC$$ just means no TOC
26-AUG-2009  MGD  $$CRUMBS$$
                  $$IF$$, $$ELSE$$ and $$ENDIF$$
                  VWCMS_SPAWN_<site-name> logical name
08-AUG-2009  MGD  fourth beta release,
                  $$TOC$CSS$..$$ pure CSS tables of content
                  $$TOC$ACROSS$$ superceded by $$TOC$CSS$$
                  AuthVmsVerifyUser() only employ ACME$_TARGET_DOI_NAME
                    when configured otherwise ACME$LATEST_ENABLED_AGENT_LIST
10-APR-2009  MGD  third beta release,
                  AuthVmsCheckIdent() and VMS rights identifiers
                  TemplateSpawn() to allow HTML-escaping of output
04-APR-2009  MGD  second beta release
                  add [include-file], [login-alias]
                  LoginAlias() to map user-supplied login name to user name
                  $$TITLE$$ additional behaviour
                  $$TOC$ACROSS$$ directive
17-MAR-2009  MGD  first beta release
24-MAR-2008  MGD  development commences
*/
/*****************************************************************************/

#define SOFTWARECR "Copyright (C) 2008-2011 Mark G.Daniel"
#define SOFTWARECD "2008-2011"

#define SOFTWAREVN "1.2.1"
#define SOFTWARENM "VWCMS"
#ifdef __ALPHA
#  define SOFTWAREID SOFTWARENM " AXP-" SOFTWAREVN
#endif
#ifdef __ia64
#  define SOFTWAREID SOFTWARENM " IA64-" SOFTWAREVN
#endif
#ifdef __VAX
#  define SOFTWAREID SOFTWARENM " VAX-" SOFTWAREVN
#endif

#define SOFTWAREGPL \
"This program, comes with ABSOLUTELY NO WARRANTY.\n\
This is free software, and you are welcome to redistribute it the under the\n\
conditions of the GNU GENERAL PUBLIC LICENSE, version 3 or any later version.\n\
http://www.gnu.org/licenses/gpl.txt\n"

#ifdef VWCMS_VMS_V62
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60200000
#undef __CRTL_VER
#define __CRTL_VER 60200000
#endif

#ifndef __VAX
#   pragma nomember_alignment
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unixlib.h>
#ifdef VAX
int decc$write_eof_to_mbx (int __fd);
#endif

/* VMS related header files */
#include <descrip.h>
#include <lib$routines.h>
#include <prvdef.h>
#include <ssdef.h>
#include <stsdef.h>
#include <syidef.h>
#include <starlet.h>

/* application specific */
#include "aes.h"
#include "cgilib.h"
#include "config.h"
#include "content.h"
#include "cookie.h"
#include "data.h"
#include "login.h"
#include "manage.h"
#include "other.h"
#include "vwcms.h"

#define FI_LI "VWCMS", __LINE__

/******************/
/* global storage */
/******************/

BOOL  CliLog,
      Debug,
      IsCgiPlus,
      RequestMethodPOST,
      WatchEnabled;

int  CgiPlusUsageCount,
     CgiPlusUsageMax = 999999999,
     PostBufferCount,
     VmsVersion;

unsigned long  CurrentSecondsTime;

unsigned long  CurrentVmsBinTime [2];
unsigned short  CurrentVmsNumTime [7];

char  *CgiEnvironmentPtr,
      *CgiHttpRefererPtr,
      *CgiHttpUserAgentPtr,
      *CgiPathInfoPtr,
      *CgiQueryStringPtr,
      *CgiRequestMethodPtr,
      *CgiRequestUriPtr,
      *CgiRemoteAddrPtr,
      *CgiRemoteHostPtr,
      *CgiScriptNamePtr,
      *CgiServerNamePtr,
      *LastModifiedNow,
      *PostBufferPtr;

char  SoftwareCopy [] = SOFTWARECR,
      SoftwareDate [] = SOFTWARECD,
      SoftwareId [] = SOFTWAREID,
      SoftwareVn [] = SOFTWAREVN,
      Utility [] = "VWCMS";

char  DebugHeader [] =
"Content-Type: text/plain\n\
Script-Control: X-content-encoding-gzip=0\n\
\n";

char  CurrentVmsTimeString [32],
      CommandLine [256],
      LinkSchemeHost [256],
      SoftwareEnv [96],
      TimeZoneString [32];

/********************/  
/* external storage */
/********************/  

extern int  ContentCacheCount,
            ContentCacheMax,
            DataCacheCount,
            DataCacheMax,
            DataCacheSizeMax;

extern CONTENT_CACHE  *ContentCache;

extern DATA_CACHE  *DataCache;

extern char  LoginItem[],
             LoginUserName[];

extern CONFIG_DATA  VWcmsConfig;

/**************/
/* prototypes */
/**************/

int sys$gettim (__unknown_params);
int sys$numtim (__unknown_params);

/*****************************************************************************/
/*
*/

main (int argc, char* argv[])
       
{
   int  status;
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (getenv("SYS$PIPE"))
   {
      /* being used inside a $ PIPE to escape HTML-forbidden characters */
      PlainPipe ();
      exit (SS$_NORMAL);
   }

   CgiLibVeeMemInUse (1);

   sprintf (SoftwareEnv, "%s, %s, %s",
            SOFTWAREID, (char*)CgiLibEnvironmentVersion(), BUILD_DATETIME);

   if (TrnLnm ("VWCMS$DBUG", NULL, 0))
   {
      /* some contortions to accomodate OSU */
      Debug = TRUE;
      if (!getenv ("WWWEXEC_RUNDOWN_STRING"))
      {
         fputs (DebugHeader, stdout);
         CgiLibEnvironmentSetDebug (1);
      }
   }

   CgiLibEnvironmentInit (argc, argv, TRUE);
   if (Debug) CgiLibEnvironmentRecordOut ();

   if (Debug)
   {
      /* continuing contortions */
      if (getenv ("WWWEXEC_RUNDOWN_STRING"))
      {
         fputs (DebugHeader, stdout);
         CgiLibEnvironmentSetDebug (0);
      }
   }

   GetParameters (argc, argv);

   if (cptr = TrnLnm ("VWCMS_USAGE_MAX", NULL, 0))
      CgiPlusUsageMax = atoi(cptr);

   CgiLibResponseSetSoftwareID (SoftwareId);
   CgiLibResponseSetErrorStatus (403);
   CgiLibResponseSetErrorMessage ("reported by VWcms");

   CgiEnvironmentPtr = CgiLibEnvironmentName ();
   VmsVersion = GetVmsVersion ();

   CgiLibResponseSetStreamMode(1);

   IsCgiPlus = CgiLibEnvironmentIsCgiPlus ();

   if (IsCgiPlus)
   {
      DataCacheInit ();

      ContentCacheInit ();

      for (;;)
      {
         /* block waiting for the next request */
         CgiLibVar ("");
         CgiPlusUsageCount++;

         if (Debug) fputs (DebugHeader, stdout);

         ProcessRequest ();

         if (Debug) CgiLibEnvironmentSetDebug (Debug = FALSE);

         if (WatchEnabled) WatchThis (FI_LI, NULL);

         CgiLibCgiPlusEOF ();

         if (CgiPlusUsageCount >= CgiPlusUsageMax) break;

         CgiLibVeeMemInit ();
      }
   }
   else
   {
      ProcessRequest ();

      if (TrnLnm ("VWCMS$WATCH_VM", NULL, 0)) CgiLibVeeMemShow (0);

      if (WatchEnabled) WatchThis (FI_LI, NULL);
   }
}

/****************************************************************************/
/*
*/ 

ProcessRequest ()

{
   int  port,
        TimeZoneSecs;
   char  *cptr, *sptr, *zptr,
         *FormSearchPtr,
         *PortPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProcessRequest()\n");

   /* initialize */
   StatTimer (FALSE);

   if (cptr = TrnLnm ("VWCMS$WATCH", NULL, 0))
   {
      if (Debug) fprintf (stdout, "VWCMS$WATCH |%s|\n", cptr);
      WatchEnabled = TRUE;
      if (isdigit(*cptr))
      {
         /* if it doesn't match the remote address then disable again */
         sptr = GetCgiVarNull ("REMOTE_ADDR");
         if (!sptr) sptr = "";
         if (strcmp (cptr, sptr)) WatchEnabled = 0;
      }
   }

   if (Debug) WatchEnabled = TRUE;

   if (WatchEnabled)
   {
      /* initialize watch mode */
      WatchThis (FI_LI, NULL);
      WatchThis (FI_LI, "SYSTEM !AZ", GetPlatformDescription());
      cptr = CgiLibVarNull ("SERVER_SOFTWARE");
      if (!cptr) cptr = "?";
      WatchThis (FI_LI, "SERVER !AZ", cptr);
      WatchThis (FI_LI, "!AZ", SoftwareId);
      if (!cptr) cptr = "?";
      WatchThis (FI_LI, "AGENT !AZ", cptr);
   }

   /* empty out any previous cookie(s) set */
   CookieSet (NULL, NULL);

   /* memory alloc by CGILIB means this *must* be done before pointers used */
   if (!strcmp (GetCgiVar("REQUEST_METHOD"), "POST"))
   {
      /********/
      /* POST */
      /********/

      if (PostBufferPtr)
      {
         /* calloc()ed by CgiLibFormRequestBody() */
         free (PostBufferPtr);
         PostBufferPtr = NULL;
         PostBufferCount = 0;
      }

      /* read and create the CGI variable equivalents of the body fields */
      CgiLibReadRequestBody (&PostBufferPtr, &PostBufferCount);
      if (WatchEnabled) WatchThis (FI_LI, "POST !UL bytes", PostBufferCount);
      if (!PostBufferPtr || !PostBufferCount) ErrorExit (SS$_BUGCHECK, FI_LI);
      if (WatchEnabled)
         WatchThis (FI_LI, "BODY !UL bytes\n!AZ", PostBufferCount, PostBufferPtr);

      CgiLibFormRequestBody (PostBufferPtr, PostBufferCount);
      if (WatchEnabled)
         while ((cptr = CgiLibVarNull ("*")) != NULL)
            WatchThis (FI_LI, "!AZ", cptr);

      /* now we can start to use the FORM_.. variables */
      RequestMethodPOST = TRUE;
   }
   else
      RequestMethodPOST = FALSE;

   /* now safe to use pointers to CGI variable values */
   if (!(CgiRemoteAddrPtr = GetCgiVarNull ("REMOTE_ADDR")))
      CgiRemoteAddrPtr = "?.?.?.?";
   if (!(CgiRemoteHostPtr = GetCgiVarNull ("REMOTE_HOST")))
      CgiRemoteHostPtr = "?.?.?.?";
   CgiServerNamePtr = GetCgiVarNull ("SERVER_NAME");
   CgiRequestMethodPtr = GetCgiVarNull ("REQUEST_METHOD");
   CgiRequestUriPtr = GetCgiVarNull ("REQUEST_URI");
   CgiScriptNamePtr = GetCgiVarNull ("SCRIPT_NAME");
   CgiPathInfoPtr = GetCgiVarNull ("PATH_INFO");
   CgiQueryStringPtr = GetCgiVarNull ("QUERY_STRING");
   CgiHttpRefererPtr = GetCgiVarNull ("HTTP_REFERER");

   /* ensure it's not empty */
   if (cptr = CgiLibVarNull ("HTTP_USER_AGENT"))
      while (*cptr && isspace(*cptr)) cptr++;
   else
      cptr = "";
   if (!*cptr) cptr = "\?\?\?";
   CgiHttpUserAgentPtr = cptr;

   port = atoi(PortPtr = GetCgiVar("SERVER_PORT"));
   zptr = (sptr = LinkSchemeHost) + sizeof(LinkSchemeHost)-1;
   cptr = GetCgiVarNull ("REQUEST_SCHEME");
   if (!cptr)
   {
      if (port == 443)
         cptr = "https:";
      else
         cptr = "http:";
   }
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   if (sptr < zptr) *sptr++ = '/';
   if (sptr < zptr) *sptr++ = '/';
   cptr = GetCgiVarNull ("HTTP_HOST");
   if (!cptr) cptr = CgiServerNamePtr;
   while (*cptr && sptr < zptr) *sptr++ = *cptr++;
   if (port && port != 80 && port != 443)
   {
      if (sptr < zptr) *sptr++ = ':';
      for (cptr = PortPtr; *cptr && sptr < zptr; *sptr++ = *cptr++);
   }
   *sptr = '\0';

   cptr = TrnLnm ("SYS$TIMEZONE_DIFFERENTIAL", NULL, 0);
   if (!cptr) cptr = TrnLnm ("VWCMS_TIMEZONE_DIFFERENTIAL", NULL, 0); 
   if (!cptr) cptr = "0";
   TimeZoneSecs = atoi(cptr);
   sprintf (TimeZoneString, "%+02.02d%02.02d",
            TimeZoneSecs/3600,
            abs((TimeZoneSecs - ((TimeZoneSecs/3600) * 3600)) / 60));
      
   sys$gettim (&CurrentVmsBinTime);
   sys$numtim (&CurrentVmsNumTime, &CurrentVmsBinTime);
   PrintFao (CurrentVmsTimeString, sizeof(CurrentVmsTimeString),
             "!20%D", &CurrentVmsBinTime);
   CurrentSecondsTime = decc$fix_time (&CurrentVmsBinTime);

   /* used to indicate dynamic page currency */
   LastModifiedNow = TimeString ("%a, %d %b %Y %H:%M:%S GMT",
                                 CurrentSecondsTime - TimeZoneSecs);

   if (cptr = TrnLnm ("VWCMS_DOWNTIME", NULL, 0))
      if (VWcmsDowntime (cptr))
         return;

   ContentProcess ();
}

/*****************************************************************************/
/*
The 'MsgPtr' string can contain a leading dollar symbol and space-delimited
host name or wildcard/regex term which is matched to the REMOTE_HOST CGI
variable.  If a match occurs the request is allowed.  If a match does not occur
the downtime message is displayed and the request terminated.  In this way a
site administrator can rework a site without general access being abailable. 
Return false to indicate the site in unavailable, false that access is allowed.
*/

BOOL VWcmsDowntime (char *MsgPtr)

{
   char  *cptr, *sptr, *zptr;
   char  HostMatch [128];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "VWcmsDowntime()\n");

   for (cptr = MsgPtr; *cptr && *cptr != ' ' && *cptr != '.'; cptr++);
   if (*MsgPtr == '$')
   {
      zptr = (sptr = HostMatch) + sizeof(HostMatch)-1;
      for (cptr = MsgPtr + 1;
           *cptr && *cptr != '$' && sptr < zptr;
           *sptr++ = *cptr++);
      *sptr = '\0';
      while (*cptr == '$') cptr++;
      if (ConfigMatch (CgiRemoteHostPtr, HostMatch)) return (FALSE);
      MsgPtr = cptr;
   }

   CgiLibResponseHeader (503, "text/html",
                         "Last-Modified: %s\n", LastModifiedNow);
   fprintf (stdout, "%s &nbsp;&nbsp;&mdash;&nbsp;&nbsp %s\n",
            RfcVmsDate(CurrentVmsTimeString), MsgPtr);

   return (TRUE);
}

/****************************************************************************/
/*
Client must be logged-in to access this page.
*/ 

void AboutVWcms ()

{
   static unsigned short  SyiHwNameLength;
   static unsigned long  SyiAvailCpuCnt,
                         SyiMemSize,
                         SyiPageSize;
   static char  SyiHwName [32],
                SyiVersion [16];
   static VMS_ITEM_LIST3 SyiItems [] = {
      { sizeof(SyiAvailCpuCnt), SYI$_AVAILCPU_CNT, &SyiAvailCpuCnt, 0 },
      { sizeof(SyiHwName)-1, SYI$_HW_NAME, &SyiHwName, &SyiHwNameLength },
      { sizeof(SyiPageSize)-1, SYI$_PAGE_SIZE, &SyiPageSize, 0 },
      { sizeof(SyiMemSize)-1, SYI$_MEMSIZE, &SyiMemSize, 0 },
      { 8, SYI$_VERSION, &SyiVersion, 0 },
      { 0,0,0,0 }
   };

   int  status,
        PageFactor;
   unsigned long  MemoryMB;
   char  *cptr,
         *RemoteHostPtr = NULL;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AboutVWcms()\n");

   CgiLibResponseHeader (200, "text/html",
"Last-Modified: %s\n\
Expires: Fri, 13 Jan 1978 14:00:00 GMT\n\
Cache-Control: no-cache, no-store, private, max-age=0, max-stale=0, \
must-revalidate, pre-check=0, post-check=0\n\
Pragma: no-cache\n%s",
                         LastModifiedNow,
                         CookieHeader());

   if (VMSnok (status = sys$getsyiw (0, 0, 0, &SyiItems, 0, 0, 0)))
      ErrorExit (status, FI_LI);

   SyiHwName[SyiHwNameLength] = '\0';
   for (cptr = SyiHwName+SyiHwNameLength-1;
        cptr > SyiHwName && isspace(*cptr);
        cptr--);
   if (cptr > SyiHwName) cptr++;
   *cptr = '\0';

   /* the number of VAX pages supported in an architecture page */
   PageFactor = SyiPageSize / 512;
#ifndef __VAX
   if (PageFactor >= 16)
      /* q&d - try and cater for systems with up to 64GB memory */
      MemoryMB = SyiMemSize * (PageFactor / 16) / 128;
   else
#endif
      MemoryMB = SyiMemSize * PageFactor / 2048;

   /* try to resolve host name from address */
   if (isdigit(*CgiRemoteHostPtr))
      RemoteHostPtr = HostNameByAddr (CgiRemoteAddrPtr);

   fprintf (stdout,
"<html>\n\
<head>\n\
<title>About VWcms</title>\n\
<style type=\"text/css\">\n\
.one { padding:1em 0 0.5em 0; font-size:120%%; text-align:left; \
text-decoration:underline; letter-spacing:1px; }\n\
.two th { padding-right:1em; text-align:left; text-decoration:underline; }\n\
.three th,td { padding-right:1em; text-align:left; white-space:nowrap; }\n\
table { border:none; } \n\
th { white-space:nowrap; margin:2px; padding-right:5px; text-align:right; font-weight:bold; } \n\
td { white-space:nowrap; margin:2px; text-align:left; } \n\
</style>\n\
</head>\n\
<html>\n\
<table>\n\
<tr><th></th><th class=\"one\">About VWcms</a></th></tr>\n\
<tr><th>Version:</td><td>%s</td></tr>\n\
<tr><th>System:</td><td>%s, %d CPU, %dMB, VMS %s</td></tr>\n\
<tr><th>Environment:</td><td>%s (%s)</td></tr>\n\
<tr><th>Server:</td><td>%s (%s)</td></tr>\n\
<tr><th>Client:</td>",
          SoftwareEnv,
          SyiHwName, SyiAvailCpuCnt, MemoryMB, SyiVersion,
          CgiLibVar("SERVER_SOFTWARE"), IsCgiPlus ? "CGIplus" : "CGI",
          CgiLibVar("SERVER_NAME"), CgiLibVar("SERVER_ADDR"));

   if (RemoteHostPtr)
      fprintf (stdout, "<td>%s (%s)</td></tr>\n",
               RemoteHostPtr, CgiRemoteAddrPtr);
   else
      fprintf (stdout, "<td>%s</td></tr>\n",
               CgiRemoteAddrPtr);

   cptr = CgiLibVarNull ("HTTP_X_FORWARDED");
   if (!cptr) cptr = CgiLibVarNull ("HTTP_FORWARDED");
   if (cptr && *cptr)
      fprintf (stdout,  "<tr><td>%s:</td><td>%s</td></tr>\n",
               "Forwarded", HtmlEscape(cptr));

   fprintf (stdout,
"<tr><th>%s:</td><td>%s</td></tr>\n\
</td></tr>\n\
<tr><td colspan=\"2\" style=\"text-align:center;padding-top:1em;\">\n\
<b>VWcms &mdash; Copyright &copy; %s Mark G. Daniel</b><br />\n\
This program, comes with ABSOLUTELY NO WARRANTY.<br />\n\
This is free software, and you are welcome to redistribute it under the \
conditions<br />\n\
of the GNU GENERAL PUBLIC LICENSE, version 3, or any later version.<br />\n\
<a href=\"http://www.gnu.org/licenses/gpl.txt\">\
http://www.gnu.org/licenses/gpl.txt</A>\n\
</td></tr>\n",
            "User-Agent", HtmlEscape(CgiHttpUserAgentPtr),
            SoftwareDate);

   if (IsCgiPlus) CgiPlusStats ();

   fputs (
"</table>\n\
</html>\n\
</body>\n", stdout);
}

/****************************************************************************/
/*
Client must be logged-in to access this page.
*/ 

void CgiPlusStats ()

{
   int  cnt, idx;
   char  *cptr;
   char  modtime [32];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CgiPlusStats()\n");

   fprintf (stdout,
"<tr><th></th><th class=\"one\">CGIplus Statistics</th></tr>\n\
<tr><th>Usage Count:</th><td>%d</td>\
<td style=\"width:100%%\"></td></tr>\n\
<tr><th>File Cache Max:</th><td>%d</td></tr>\n\
<tr><th>File Cache Count:</th><td>%d</td></tr>\n\
<tr><th>File Cache Size Max:</th><td>%d&nbsp;&nbsp;kBytes</td></tr>\n\
<tr><th>Content Cache Max:</th><td>%d</td></tr>\n\
<tr><th>Content Cache Count:</th><td>%d</td></tr>\n\
</td></tr>\n",
            CgiPlusUsageCount,
            DataCacheMax,
            DataCacheCount,
            DataCacheSizeMax,
            ContentCacheMax,
            ContentCacheCount);

   if (DataCacheCount)
   {
      fprintf (stdout,
"<tr><th></th><th class=\"one\">File Cache</th></tr>\n\
<tr><td colspan=\"2\">\n\
<table border=\"0\" cellpadding=\"0\" cellspacing=\"5\">\n\
<tr class=\"two\">\
<th></th><th style=\"text-align:left;\">File Name</th>\
<th style=\"text-align:left;\">Modified</th>\
<th>Bytes</th><th>Hit</th><th>Reused</th></tr>\n");

      for (idx = cnt = 0; idx < DataCacheCount; idx++)
      {
         if (!DataCache[idx].FileName[0]) continue;

         strcpy (modtime, ctime((unsigned long*)&DataCache[idx].mtime));
         for (cptr = modtime; *cptr && *cptr != '\n'; cptr++);
         *cptr = '\0';

         fprintf (stdout,
"<tr class=\"three\">\
<th style=\"text-align:right;\">%d.</th>\
<td style=\"text-align:left; font-family:monospace;\">%s</td>\
<td style=\"text-align:left;\">%s</td>\
<td>%d</td>\
<td>%d</td>\
<td>%d</td></tr>\n",
                  ++cnt,
                  DataCache[idx].FileName,
                  modtime,
                  DataCache[idx].ContentLength,
                  DataCache[idx].HitCount,
                  DataCache[idx].ReusedCount);
      }

      fputs (
"</table>\n\
</td></tr>\n",
             stdout);
   }

   if (ContentCacheCount)
   {
      fprintf (stdout,
"<tr><th></th><th class=\"one\">Content Cache</th></tr>\n\
<tr><td colspan=\"2\">\n\
<table border=\"0\" cellpadding=\"0\" cellspacing=\"5\">\n\
<tr class=\"two\">\
<th></th><th style=\"text-align:left;\">File Name</th>\
<th style=\"text-align:left;\">Modified</th>\
<th>Bytes</th><th>Hit</th><th>Reused</th></tr>\n");

      for (idx = cnt = 0; idx < ContentCacheCount; idx++)
      {
         if (!ContentCache[idx].FileName[0]) continue;

         strcpy (modtime, ctime((unsigned long*)&ContentCache[idx].mtime));
         for (cptr = modtime; *cptr && *cptr != '\n'; cptr++);
         *cptr = '\0';

         fprintf (stdout,
"<tr class=\"three\">\
<th style=\"text-align:right;\">%d.</th>\
<td style=\"text-align:left; font-family:monospace;\">%s</td>\
<td style=\"text-align:left;\">%s</td>\
<td>%d</td>\
<td>%d</td>\
<td>%d</td></tr>\n",
                  ++cnt,
                  ContentCache[idx].FileName,
                  modtime,
                  ContentCache[idx].TextLength,
                  ContentCache[idx].HitCount,
                  ContentCache[idx].ReusedCount);
      }

      fputs (
"</table>\n\
</td></tr>\n",
             stdout);
   }

   fprintf (stdout,
"<tr><th></th><th class=\"one\">Virtual Memory</th></tr>\n\
<tr><td colspan=\"2\"><pre>");

   CgiLibVeeMemShow (NULL);

   fputs (
"</pre>\n\
</td></tr>\n\
</table>\n\
\
</td></tr>\n",
          stdout);
}

/****************************************************************************/
/*
This function is intended to be used inside a VMS DCL PIPE to escape
HTML-forbidden characters and add HTML line-breaks ("<br />") inside
$$INCLUDE$$ spawn commands as in "pipe show process/all | run cgi_exe:vwcms".
*/ 

void PlainPipe ()

{
   int  len;
   char  *cptr, *sptr, *zptr;
   char  PipeBuffer [2048],
         EscapeBuffer [4096];
   FILE  *SysPipe;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "PlainPipe()\n");

   if (!(stdout = freopen ("SYS$OUTPUT:", "w", stdout, "ctx=bin")))
      exit (vaxc$errno);

   if (!(SysPipe = fopen ("SYS$PIPE", "r", "ctx=rec")))
      exit (vaxc$errno);

   while ((len = decc$record_read (SysPipe, PipeBuffer, sizeof(PipeBuffer))) > 0)
   {
      PipeBuffer[len] = '\0';
      /* allow elbow-room for worst-case */
      zptr = (sptr = EscapeBuffer) + sizeof(EscapeBuffer)-16;
      for (cptr = PipeBuffer; *cptr && sptr < zptr; cptr++)
      {
         switch (*cptr)
         {
            case '<' :
               *sptr++ = '&';
               *sptr++ = 'l';
               *sptr++ = 't';
               *sptr++ = ';';
               break;
            case '>' :
               *sptr++ = '&';
               *sptr++ = 'g';
               *sptr++ = 't';
               *sptr++ = ';';
               break;
            case '&' :
               *sptr++ = '&';
               *sptr++ = 'a';
               *sptr++ = 'm';
               *sptr++ = 'p';
               *sptr++ = ';';
               break;
            default :
               *sptr++ = *cptr;
         }
      }
      for (cptr = "<br />"; *cptr && sptr < zptr; *sptr++ = *cptr++);
      *sptr = '\0';
      decc$record_write (stdout, EscapeBuffer, sptr-EscapeBuffer);
   }

   decc$write_eof_to_mbx (fileno(stdout));
   fclose (SysPipe);
}

/****************************************************************************/
/*
*/ 

void CancelCookies ()

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CancelCookies()\n");

   CookieSet (NULL, NULL);
   CookieSet (LOGIN_SESSION_COOKIE, NULL);
   CookieSet (MANAGE_COOKIE, NULL);

   CgiLibResponseHeader (200, "text/plain",
"Last-Modified: %s\n\
Expires: Fri, 13 Jan 1978 14:00:00 GMT\n\
Cache-Control: no-cache, no-store, private, max-age=0, max-stale=0, \
must-revalidate, pre-check=0, post-check=0\n\
Pragma: no-cache\n%s",
                         LastModifiedNow,
                         CookieHeader());

   fputs ("Cookies reset!\n", stdout);
}

/*****************************************************************************/
/*
Get "command-line" parameters, whether from the command-line or from a
configuration symbol or logical containing the equivalent.
*/

GetParameters
(
int argc,
char *argv[]
)
{
   static unsigned long  Flags = 0;

   int  status,
        SkipParameters;
   unsigned short  Length;
   char  ch;
   char  *aptr, *cptr, *clptr, *sptr;
   $DESCRIPTOR (CommandLineDsc, CommandLine);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "GetParameters()\n");

   if ((clptr = getenv ("VWCMS$PARAM")) == NULL)
   {
      /* get the entire command line following the verb */
      status = lib$get_foreign (&CommandLineDsc, 0, &Length, &Flags);
      if (VMSnok (status)) exit (status);
      (clptr = CommandLine)[Length] = '\0';
   }

   /* the first is CSWS 1.2/1.3, the other two for SWS 2.0 */
   if (getenv ("APACHE$SHARED_SOCKET") != NULL ||
       getenv ("APACHE$PARENT_PID") != NULL ||
       getenv ("APACHE$SERVER_TAG") != NULL)
      SkipParameters = 99;
   else
   /* if OSU environment then skip P1, P2, P3 */
   if (getenv ("WWWEXEC_RUNDOWN_STRING") != NULL)
      SkipParameters = 3;
   else
      SkipParameters = 0;

   aptr = NULL;
   ch = *clptr;
   for (;;)
   {
      if (aptr != NULL && *aptr == '/') *aptr = '\0';
      if (!ch) break;

      *clptr = ch;
      if (Debug) fprintf (stdout, "clptr |%s|\n", clptr);
      while (*clptr && isspace(*clptr)) *clptr++ = '\0';
      aptr = clptr;
      if (*clptr == '/') clptr++;
      while (*clptr && !isspace (*clptr) && *clptr != '/')
      {
         if (*clptr != '\"')
         {
            clptr++;
            continue;
         }
         cptr = clptr;
         clptr++;
         while (*clptr)
         {
            if (*clptr == '\"')
               if (*(clptr+1) == '\"')
                  clptr++;
               else
                  break;
            *cptr++ = *clptr++;
         }
         *cptr = '\0';
         if (*clptr) clptr++;
      }
      ch = *clptr;
      if (*clptr) *clptr = '\0';
      if (Debug) fprintf (stdout, "aptr |%s|\n", aptr);
      if (!*aptr) continue;

      if (SkipParameters)
      {
         /*
             Apache supplies query string keywords on the command line
             to a CGI script.  With Apache the command-line is ignored
             up until the /_APACHE qualifier triggers anything following
             to be processed.             
         */
         if (!strsame (aptr, "/_APACHE", -1))
         {
            SkipParameters--;
            continue;
         }
         SkipParameters = 0;
      }

      /***********/
      /* process */
      /***********/

      if (strsame (aptr, "/AES_SELF_TEST", -1))
      {
         aes_self_test (1);
         exit (SS$_NORMAL);
      }

      if (strsame (aptr, "/DBUG", -1))
      {
         Debug = TRUE;
         continue;
      }

      if (strsame (aptr, "/VERSION", 4))
      {
         fprintf (stdout, "%%%s-I-VERSION, %s\n%s\n%s",
                  Utility, SoftwareEnv, SOFTWARECR, SOFTWAREGPL);
         exit (SS$_NORMAL);
      }

      if (strsame (aptr, "/WATCH", 4))
      {
         WatchEnabled = TRUE;
         continue;
      }

      if (*aptr == '/')
      {
         fprintf (stdout, "%%%s-E-IVQUAL, unrecognized qualifier\n \\%s\\\n",
                  Utility, aptr+1);
         exit (STS$K_ERROR | STS$M_INHIB_MSG);
      }
   }
}

/*****************************************************************************/

